"""Abstract base for simulation data structures."""

from __future__ import annotations

import pathlib
from abc import ABC
from os import PathLike
from typing import Any, Optional, Union

import numpy as np
import pydantic.v1 as pd
import xarray as xr

from tidy3d.components.base import Tidy3dBaseModel, skip_if_fields_missing
from tidy3d.components.base_sim.data.monitor_data import AbstractMonitorData
from tidy3d.components.base_sim.simulation import AbstractSimulation
from tidy3d.components.data.utils import UnstructuredGridDatasetType
from tidy3d.components.file_util import replace_values
from tidy3d.components.monitor import AbstractMonitor
from tidy3d.components.types import FieldVal
from tidy3d.exceptions import DataError, FileError, Tidy3dKeyError, ValidationError


class AbstractSimulationData(Tidy3dBaseModel, ABC):
    """Stores data from a collection of :class:`AbstractMonitor` objects in
    a :class:`AbstractSimulation`.
    """

    simulation: AbstractSimulation = pd.Field(
        ...,
        title="Simulation",
        description="Original :class:`AbstractSimulation` associated with the data.",
    )

    data: tuple[AbstractMonitorData, ...] = pd.Field(
        ...,
        title="Monitor Data",
        description="List of :class:`AbstractMonitorData` instances "
        "associated with the monitors of the original :class:`AbstractSimulation`.",
    )

    log: str = pd.Field(
        None,
        title="Solver Log",
        description="A string containing the log information from the simulation run.",
    )

    def __getitem__(self, monitor_name: str) -> AbstractMonitorData:
        """Get a :class:`.AbstractMonitorData` by name. Apply symmetry if applicable."""
        monitor_data = self.monitor_data[monitor_name]
        return monitor_data.symmetry_expanded_copy

    @property
    def monitor_data(self) -> dict[str, AbstractMonitorData]:
        """Dictionary mapping monitor name to its associated :class:`AbstractMonitorData`."""
        return {monitor_data.monitor.name: monitor_data for monitor_data in self.data}

    @pd.root_validator(skip_on_failure=True)
    def data_monitors_match_sim(cls, values):
        """Ensure each :class:`AbstractMonitorData` in ``.data`` corresponds to a monitor in
        ``.simulation``.
        """
        sim = values.get("simulation")
        data = values.get("data")

        for mnt_data in data:
            try:
                monitor_name = mnt_data.monitor.name
                sim.get_monitor_by_name(monitor_name)
            except Tidy3dKeyError as exc:
                raise DataError(
                    f"Data with monitor name '{monitor_name}' supplied "
                    f"but not found in the original '{sim.type}'."
                ) from exc
        return values

    @pd.validator("data", always=True)
    @skip_if_fields_missing(["simulation"])
    def validate_no_ambiguity(cls, val, values):
        """Ensure all :class:`AbstractMonitorData` entries in ``.data`` correspond to different
        monitors in ``.simulation``.
        """
        names = [mnt_data.monitor.name for mnt_data in val]

        if len(set(names)) != len(names):
            raise ValidationError("Some entries of '.data' provide data for same monitor(s).")

        return val

    @staticmethod
    def _field_component_value(
        field_component: Union[xr.DataArray, UnstructuredGridDatasetType], val: FieldVal
    ) -> xr.DataArray:
        """return the desired value of a field component.

        Parameter
        ----------
        field_component : Union[xarray.DataArray, UnstructuredGridDatasetType]
            Field component from which to calculate the value.
        val : Literal['real', 'imag', 'abs', 'abs^2', 'phase']
            Which part of the field to return.

        Returns
        -------
        xarray.DataArray
            Value extracted from the field component.
        """
        if val in ("real", "re"):
            field_value = field_component.real
            field_value = field_value.rename(f"Re{{{field_component.name}}}")

        elif val in ("imag", "im"):
            field_value = field_component.imag
            field_value = field_value.rename(f"Im{{{field_component.name}}}")

        elif val == "abs":
            field_value = np.abs(field_component)
            field_value = field_value.rename(f"|{field_component.name}|")

        elif val == "abs^2":
            field_value = np.abs(field_component) ** 2
            field_value = field_value.rename(f"|{field_component.name}|²")

        elif val == "phase":
            field_value = np.arctan2(field_component.imag, field_component.real)
            field_value = field_value.rename(f"∠{field_component.name}")

        else:
            raise Tidy3dKeyError(
                f"Couldn't find 'val={val}'. Must be one of 'real', 're', 'imag', 'im', 'abs', 'abs^2', 'phase'."
            )

        return field_value

    @staticmethod
    def _apply_log_scale(
        field_data: xr.DataArray,
        vmin: Optional[float] = None,
        db_factor: float = 1.0,
    ) -> xr.DataArray:
        """Prepare field data for log-scale plotting by handling zeros.

        Takes absolute value of the data, replaces zeros with a fill value
        (to prevent log10(0) warnings), and applies log10 scaling.

        Parameters
        ----------
        field_data : xr.DataArray
            The field data to prepare.
        vmin : float, optional
            The minimum value for the color scale. If provided, zeros are replaced
            with ``10 ** (vmin / db_factor)`` instead of NaN.
        db_factor : float
            Factor to multiply the log10 result by (e.g., 20 for dB scale of field,
            10 for dB scale of power). Default is 1 (pure log10 scale).

        Returns
        -------
        xr.DataArray
            The log-scaled field data.
        """
        fill_val = np.nan
        if vmin is not None:
            fill_val = 10 ** (vmin / db_factor)
        field_data = np.abs(field_data)
        field_data = field_data.where((field_data > 0) | np.isnan(field_data), fill_val)
        return db_factor * np.log10(field_data)

    def get_monitor_by_name(self, name: str) -> AbstractMonitor:
        """Return monitor named 'name'."""
        return self.simulation.get_monitor_by_name(name)

    def to_mat_file(self, fname: PathLike, **kwargs: Any) -> None:
        """Output the simulation data object as ``.mat`` MATLAB file.

        Parameters
        ----------
        fname : PathLike
            Full path to the output file. Should include ``.mat`` file extension.
        **kwargs : dict, optional
            Extra arguments to ``scipy.io.savemat``: see ``scipy`` documentation for more detail.

        Example
        -------
        >>> sim_data.to_mat_file('/path/to/file/data.mat') # doctest: +SKIP
        """
        # Check .mat file extension is given
        extension = pathlib.Path(fname).suffixes[0].lower()
        if len(extension) == 0:
            raise FileError(f"File '{fname}' missing extension.")
        if extension != ".mat":
            raise FileError(f"File '{fname}' should have a .mat extension.")

        # Handle m_dict in kwargs
        if "m_dict" in kwargs:
            raise ValueError(
                "'m_dict' is automatically determined by 'to_mat_file', can't pass to 'savemat'."
            )

        # Get SimData object as dictionary
        sim_dict = self.dict()

        # set long field names true by default, otherwise it wont save fields with > 31 characters
        if "long_field_names" not in kwargs:
            kwargs["long_field_names"] = True

        # Remove NoneType values from dict
        # Built from theory discussed in https://github.com/scipy/scipy/issues/3488
        modified_sim_dict = replace_values(sim_dict, None, [])

        try:
            from scipy.io import savemat

            savemat(fname, modified_sim_dict, **kwargs)
        except Exception as e:
            raise ValueError(
                "Could not save supplied simulation data to file. As this is an experimental "
                "feature, we may not be able to support the contents of your dataset. If you "
                "receive this error, please feel free to raise an issue on our front end "
                "repository so we can investigate."
            ) from e



from __future__ import annotations

from collections.abc import Sequence

from numpy.typing import NDArray


def symmetrize_mirror(array: NDArray, axis: int | tuple[int, int]) -> NDArray:
    """
    Symmetrizes the parameter array by averaging the mirrored parts of the array. The axis argument specifies the
    symmetry axis.

    Parameters
    ----------
    array : NDArray
        The input array to be symmetrized.
    axis: int | tuple[int, int]
        The symmetry axis. This can either be a single axis (x=0 or y=1) or a tuple of both axes.

    Returns
    -------
    NDArray
        The array after applying the mirror symmetrization.

    Example
    -------
        >>> import autograd.numpy as np
        >>> from tidy3d.plugins.autograd.invdes.symmetries import symmetrize_mirror
        >>> arr = np.asarray([
        ...     [1, 2],
        ...     [3, 4]
        ... ])
        >>> res_x = symmetrize_mirror(arr, axis=0)
        >>> res_y = symmetrize_mirror(arr, axis=1)
        >>> assert np.all(np.isclose(res_x, np.asarray([
        ...     [2, 3],
        ...     [2, 3]
        ... ])))
        >>> assert np.all(np.isclose(res_y, np.asarray([
        ...     [1.5, 1.5],
        ...     [3.5, 3.5]
        ... ])))
    """
    if isinstance(axis, int) and (axis != 0 and axis != 1):
        raise ValueError(f"Invalid axis: expected 0, 1, or Sequence thereof, but got {axis}")
    if array.ndim != 2:
        raise ValueError(f"Invalid array shape: {array.shape}. Need 2d array.")
    if isinstance(axis, Sequence) and (
        len(axis) != 2 or axis[0] not in [0, 1] or axis[1] not in [0, 1] or axis[0] == axis[1]
    ):
        raise ValueError(f"Invalid axis: expected 0, 1, or Sequence thereof, but got {axis}")

    # Helper function to flip along a specific axis using slicing
    # Autograd supports slicing (e.g. ::-1) but lacks VJP for np.flip
    def flip_axis(arr, ax):
        if ax == 0:
            return arr[::-1, :]
        elif ax == 1:
            return arr[:, ::-1]
        return arr

    # Case 1: Symmetrize along both axes
    if isinstance(axis, Sequence):
        # Symmetrize along axis 0
        array = (array + flip_axis(array, 0)) / 2.0
        # Symmetrize along axis 1
        array = (array + flip_axis(array, 1)) / 2.0
        return array

    # Case 2: Symmetrize along a single axis
    return (array + flip_axis(array, axis)) / 2.0


def symmetrize_rotation(array: NDArray) -> NDArray:
    """
    Symmetrizes the parameter array by averaging over all four 90-degree rotations.
    The input array must be square.

    Parameters
    ----------
    array : NDArray
        The input array to be symmetrized.

    Returns
    -------
    NDArray
        The array after applying rotational symmetrization.

    Example
    -------
        >>> import autograd.numpy as np
        >>> from tidy3d.plugins.autograd.invdes.symmetries import symmetrize_rotation
        >>> arr = np.asarray([
        ...     [0, 0, 0],
        ...     [1, 5, 2],
        ...     [8, 1, 8]
        ... ])
        >>> res = symmetrize_rotation(arr)
        >>> assert np.all(np.isclose(res, np.asarray([
        ...     [4, 1, 4],
        ...     [1, 5, 1],
        ...     [4, 1, 4]
        ... ])))
    """
    if array.ndim != 2:
        raise ValueError(f"Invalid array shape: {array.shape}. Need 2d array.")

    if array.shape[0] != array.shape[1]:
        raise ValueError(
            f"Invalid array shape: {array.shape}. Array must be square for rotational symmetry."
        )

    # Average the array with its 90, 180, and 270 degree rotations
    # We manually implement rotations using Transpose (.T) and Slicing ([::-1])
    # to ensure full compatibility with autograd.

    # 0 degrees: array
    # 90 degrees CCW: Transpose -> Flip Rows (equivalent to array.T[::-1, :])
    rot90 = array.T[::-1, :]

    # 180 degrees: Flip Rows -> Flip Cols (equivalent to array[::-1, ::-1])
    rot180 = array[::-1, ::-1]

    # 270 degrees CCW: Transpose -> Flip Cols (equivalent to array.T[:, ::-1])
    rot270 = array.T[:, ::-1]

    return (array + rot90 + rot180 + rot270) / 4.0


def symmetrize_diagonal(array: NDArray, anti: bool = False) -> NDArray:
    """
    Symmetrizes the parameter array by averaging it with its transpose.
    Can symmetrize along the main diagonal or the anti-diagonal.

    Parameters
    ----------
    array : NDArray
        The input array to be symmetrized. Must be square.
    anti : bool, optional
        If False (default), symmetrizes along the main diagonal (top-left to bottom-right).
        If True, symmetrizes along the anti-diagonal (top-right to bottom-left).

    Returns
    -------
    NDArray
        The array after applying diagonal symmetrization.

    Example
    -------
        >>> import autograd.numpy as np
        >>> from tidy3d.plugins.autograd.invdes.symmetries import symmetrize_diagonal
        >>> arr = np.asarray([
        ...     [1, 2],
        ...     [4, 7],
        ... ])
        >>> res = symmetrize_diagonal(arr)
        >>> res_anti = symmetrize_diagonal(arr, anti=True)
        >>> assert np.all(np.isclose(res, np.asarray([
        ...     [1, 3],
        ...     [3, 7]
        ... ])))
        >>> assert np.all(np.isclose(res_anti, np.asarray([
        ...     [4, 2],
        ...     [4, 4]
        ... ])))
    """
    if array.ndim != 2:
        raise ValueError(f"Invalid array shape: {array.shape}. Need 2d array.")

    if array.shape[0] != array.shape[1]:
        raise ValueError(
            f"Invalid array shape: {array.shape}. Array must be square for diagonal symmetry."
        )

    if anti:
        # Anti-diagonal symmetrization.
        # Mathematically equivalent to averaging A with the flip of A transposed.
        # np.flip(array.T) is equivalent to array.T[::-1, ::-1]
        anti_transpose = array.T[::-1, ::-1]
        return (array + anti_transpose) / 2.0

    # Standard main diagonal symmetry: (A + A_transpose) / 2
    return (array + array.T) / 2.0



"""Type definitions for monitor data."""

from __future__ import annotations

from typing import Union

from tidy3d.components.data.monitor_data import (
    AuxFieldTimeData,
    DiffractionData,
    DirectivityData,
    FieldData,
    FieldProjectionAngleData,
    FieldProjectionCartesianData,
    FieldProjectionKSpaceData,
    FieldTimeData,
    FluxData,
    FluxTimeData,
    MediumData,
    ModeData,
    ModeSolverData,
    PermittivityData,
)
from tidy3d.components.microwave.data.monitor_data import MicrowaveModeData, MicrowaveModeSolverData

# Type aliases
ModeDataType = Union[ModeData, MicrowaveModeData]
ModeSolverDataType = Union[ModeSolverData, MicrowaveModeSolverData]
MonitorDataTypes = (
    FieldData,
    FieldTimeData,
    PermittivityData,
    MediumData,
    ModeSolverData,
    ModeData,
    FluxData,
    FluxTimeData,
    AuxFieldTimeData,
    FieldProjectionKSpaceData,
    FieldProjectionCartesianData,
    FieldProjectionAngleData,
    DiffractionData,
    DirectivityData,
    MicrowaveModeData,
    MicrowaveModeSolverData,
)
MonitorDataType = Union[MonitorDataTypes]



"""Backwards compatibility - import from tidy3d.components.microwave.path_integrals.integrals instead."""

from __future__ import annotations

from tidy3d.components.microwave.path_integrals.integrals.base import (
    AxisAlignedPathIntegral,
    EMScalarFieldType,
    IntegrableMonitorDataType,
)
from tidy3d.components.microwave.path_integrals.integrals.current import (
    AxisAlignedCurrentIntegral,
)
from tidy3d.components.microwave.path_integrals.integrals.voltage import (
    AxisAlignedVoltageIntegral,
)

__all__ = [
    "AxisAlignedCurrentIntegral",
    "AxisAlignedPathIntegral",
    "AxisAlignedVoltageIntegral",
    "EMScalarFieldType",
    "IntegrableMonitorDataType",
]



"""Imports from microwave plugin."""

from __future__ import annotations

from tidy3d.components.microwave.impedance_calculator import (
    CurrentIntegralType,
    ImpedanceCalculator,
    VoltageIntegralType,
)
from tidy3d.components.microwave.path_integrals.integrals.auto import (
    path_integrals_from_lumped_element,
)
from tidy3d.components.microwave.path_integrals.integrals.base import (
    AxisAlignedPathIntegral,
    Custom2DPathIntegral,
)
from tidy3d.components.microwave.path_integrals.integrals.current import (
    AxisAlignedCurrentIntegral,
    CompositeCurrentIntegral,
    Custom2DCurrentIntegral,
)
from tidy3d.components.microwave.path_integrals.integrals.voltage import (
    AxisAlignedVoltageIntegral,
    Custom2DVoltageIntegral,
)

from . import models
from .array_factor import (
    BlackmanHarrisWindow,
    BlackmanWindow,
    ChebWindow,
    HammingWindow,
    HannWindow,
    KaiserWindow,
    RadialTaper,
    RectangularAntennaArrayCalculator,
    RectangularTaper,
    TaylorWindow,
)
from .lobe_measurer import LobeMeasurer
from .rf_material_library import rf_material_library

# Backwards compatibility: Remove 2.11 renamed integral classes
CurrentIntegralTypes = CurrentIntegralType
VoltageIntegralTypes = VoltageIntegralType
VoltageIntegralAxisAligned = AxisAlignedVoltageIntegral
CurrentIntegralAxisAligned = AxisAlignedCurrentIntegral
CustomPathIntegral2D = Custom2DPathIntegral
CustomVoltageIntegral2D = Custom2DVoltageIntegral
CustomCurrentIntegral2D = Custom2DCurrentIntegral

__all__ = [
    "AxisAlignedCurrentIntegral",
    "AxisAlignedPathIntegral",
    "AxisAlignedVoltageIntegral",
    "BlackmanHarrisWindow",
    "BlackmanWindow",
    "ChebWindow",
    "CompositeCurrentIntegral",
    "CurrentIntegralAxisAligned",  # Backwards compatibility alias
    "CurrentIntegralTypes",
    "Custom2DCurrentIntegral",
    "Custom2DPathIntegral",
    "Custom2DVoltageIntegral",
    "CustomCurrentIntegral2D",  # Backwards compatibility alias
    "CustomPathIntegral2D",  # Backwards compatibility alias
    "CustomVoltageIntegral2D",  # Backwards compatibility alias
    "HammingWindow",
    "HannWindow",
    "ImpedanceCalculator",
    "KaiserWindow",
    "LobeMeasurer",
    "RadialTaper",
    "RectangularAntennaArrayCalculator",
    "RectangularTaper",
    "TaylorWindow",
    "VoltageIntegralAxisAligned",  # Backwards compatibility alias
    "VoltageIntegralTypes",
    "models",
    "path_integrals_from_lumped_element",
    "rf_material_library",
]



from __future__ import annotations

from collections.abc import Iterable
from typing import Any, get_args, get_origin

import tomlkit
from pydantic import BaseModel
from pydantic.fields import FieldInfo
from tomlkit.items import Item, Table

from .registry import get_sections

Path = tuple[str, ...]


def collect_descriptions() -> dict[Path, str]:
    """Collect description strings for registered configuration fields."""

    descriptions: dict[Path, str] = {}
    for section_name, model in get_sections().items():
        base_path = tuple(segment for segment in section_name.split(".") if segment)
        section_doc = (model.__doc__ or "").strip()
        if section_doc and base_path:
            descriptions[base_path] = descriptions.get(
                base_path, section_doc.splitlines()[0].strip()
            )
        for field_name, field in model.model_fields.items():
            descriptions.update(_describe_field(field, prefix=(*base_path, field_name)))
    return descriptions


def _describe_field(field: FieldInfo, prefix: Path) -> dict[Path, str]:
    descriptions: dict[Path, str] = {}
    description = (field.description or "").strip()
    if description:
        descriptions[prefix] = description

    nested_models: Iterable[type[BaseModel]] = _iter_model_types(field.annotation)
    for model in nested_models:
        nested_doc = (model.__doc__ or "").strip()
        if nested_doc:
            descriptions[prefix] = descriptions.get(prefix, nested_doc.splitlines()[0].strip())
        for sub_name, sub_field in model.model_fields.items():
            descriptions.update(_describe_field(sub_field, prefix=(*prefix, sub_name)))
    return descriptions


def _iter_model_types(annotation: Any) -> Iterable[type[BaseModel]]:
    """Yield BaseModel subclasses referenced by a field annotation (if any)."""

    if annotation is None:
        return

    stack = [annotation]
    seen: set[type[BaseModel]] = set()

    while stack:
        current = stack.pop()
        if isinstance(current, type) and issubclass(current, BaseModel):
            if current not in seen:
                seen.add(current)
                yield current
            continue

        origin = get_origin(current)
        if origin is None:
            continue

        stack.extend(get_args(current))


def build_document(
    data: dict[str, Any],
    existing: tomlkit.TOMLDocument | None,
    descriptions: dict[Path, str] | None = None,
) -> tomlkit.TOMLDocument:
    """Return a TOML document populated with data and annotated comments."""

    descriptions = descriptions or collect_descriptions()
    document = existing if existing is not None else tomlkit.document()
    _prune_missing_keys(document, data.keys())
    for key, value in data.items():
        _apply_value(
            container=document,
            key=key,
            value=value,
            path=(key,),
            descriptions=descriptions,
            is_new=key not in document,
        )
    return document


def _prune_missing_keys(container: Table | tomlkit.TOMLDocument, keys: Iterable[str]) -> None:
    desired = set(keys)
    for existing_key in list(container.keys()):
        if existing_key not in desired:
            del container[existing_key]


def _apply_value(
    container: Table | tomlkit.TOMLDocument,
    key: str,
    value: Any,
    path: Path,
    descriptions: dict[Path, str],
    is_new: bool,
) -> None:
    description = descriptions.get(path)
    if isinstance(value, dict):
        existing = container.get(key)
        table = existing if isinstance(existing, Table) else tomlkit.table()
        _prune_missing_keys(table, value.keys())
        for sub_key, sub_value in value.items():
            _apply_value(
                container=table,
                key=sub_key,
                value=sub_value,
                path=(*path, sub_key),
                descriptions=descriptions,
                is_new=not isinstance(existing, Table) or sub_key not in table,
            )
        if key in container:
            container[key] = table
        else:
            if isinstance(container, tomlkit.TOMLDocument) and len(container) > 0:
                container.add(tomlkit.nl())
            container.add(key, table)
        return

    if value is None:
        return

    existing_item = container.get(key)
    new_item = tomlkit.item(value)
    if isinstance(existing_item, Item):
        new_item.trivia.comment = existing_item.trivia.comment
        new_item.trivia.comment_ws = existing_item.trivia.comment_ws
    elif description:
        new_item.comment(description)

    if key in container:
        container[key] = new_item
    else:
        container.add(key, new_item)



"""Abstract base class for all ports in the component and terminal component modelers."""

from __future__ import annotations

from abc import ABC

import pydantic.v1 as pd

from tidy3d.components.base import Tidy3dBaseModel
from tidy3d.exceptions import SetupError


class AbstractBasePort(Tidy3dBaseModel, ABC):
    """Abstract base class representing a port excitation of a component."""

    name: str = pd.Field(
        ...,
        title="Name",
        description="Unique name for the port.",
        min_length=1,
    )

    @pd.validator("name")
    def _valid_port_name(cls, val):
        """Make sure port name does not include the '@' symbol, so that task names will always be unique."""
        if "@" in val:
            raise SetupError(f"Port names must not include the '@' symbol. Name given was '{val}'.")
        return val



"""Built-in configuration profiles for tidy3d."""

from __future__ import annotations

from typing import Any

BUILTIN_PROFILES: dict[str, dict[str, Any]] = {
    "default": {
        "web": {
            "api_endpoint": "https://tidy3d-api.simulation.cloud",
            "website_endpoint": "https://tidy3d.simulation.cloud",
            "s3_region": "us-gov-west-1",
        }
    },
    "prod": {
        "web": {
            "api_endpoint": "https://tidy3d-api.simulation.cloud",
            "website_endpoint": "https://tidy3d.simulation.cloud",
            "s3_region": "us-gov-west-1",
        }
    },
    "dev": {
        "web": {
            "api_endpoint": "https://tidy3d-api.dev-simulation.cloud",
            "website_endpoint": "https://tidy3d.dev-simulation.cloud",
            "s3_region": "us-east-1",
        }
    },
    "uat": {
        "web": {
            "api_endpoint": "https://tidy3d-api.uat-simulation.cloud",
            "website_endpoint": "https://tidy3d.uat-simulation.cloud",
            "s3_region": "us-west-2",
        }
    },
    "pre": {
        "web": {
            "api_endpoint": "https://preprod-tidy3d-api.simulation.cloud",
            "website_endpoint": "https://preprod-tidy3d.simulation.cloud",
            "s3_region": "us-gov-west-1",
        }
    },
    "nexus": {
        "web": {
            "api_endpoint": "http://127.0.0.1:5000",
            "website_endpoint": "http://127.0.0.1/tidy3d",
            "ssl_verify": False,
            "enable_caching": False,
            "s3_region": "us-east-1",
            "env_vars": {
                "AWS_ENDPOINT_URL_S3": "http://127.0.0.1:9000",
            },
        }
    },
}

__all__ = ["BUILTIN_PROFILES"]



# transformations applied to design region
from __future__ import annotations

import abc
import typing
from typing import Any

import autograd.numpy as anp
import pydantic.v1 as pd

import tidy3d as td
from tidy3d.plugins.autograd.functions import threshold
from tidy3d.plugins.autograd.invdes import make_filter_and_project

from .base import InvdesBaseModel


class AbstractTransformation(InvdesBaseModel, abc.ABC):
    """Base class for transformations."""

    @abc.abstractmethod
    def evaluate(self) -> float:
        """Evaluate the penalty on supplied values."""

    def __call__(self, *args: Any, **kwargs: Any) -> float:
        return self.evaluate(*args, **kwargs)


class FilterProject(InvdesBaseModel):
    """Transformation involving convolution by a conic filter followed by a ``tanh`` projection.

    Notes
    -----

        .. image:: ../../_static/img/filter_project.png

    """

    radius: pd.PositiveFloat = pd.Field(
        ...,
        title="Filter Radius",
        description="Radius of the filter to convolve with supplied spatial data. "
        "Note: the corresponding feature size expressed in the device is typically "
        "sqrt(3) times smaller than the radius. For best results, it is recommended to make your "
        "radius about twice as large as the desired feature size. "
        "Note: the radius value is often only approximately related to the final feature sizes. "
        "It is useful to apply a ``FilterProject`` transformation to 'encourage' larger "
        "feature sizes, but we ultimately recommend creating a ``ErosionDilationPenalty`` to the "
        "``DesignRegion.penalties`` if you have strict fabrication constraints.",
        units=td.constants.MICROMETER,
    )

    beta: float = pd.Field(
        1.0,
        ge=1.0,
        title="Beta",
        description="Steepness of the binarization, "
        "higher means more sharp transition "
        "at the expense of gradient accuracy and ease of optimization. ",
    )

    eta: float = pd.Field(
        0.5, ge=0.0, le=1.0, title="Eta", description="Halfway point in projection function."
    )

    strict_binarize: bool = pd.Field(
        False,
        title="Binarize strictly",
        description="If ``False``, the binarization is still continuous between min and max. "
        "If ``True``, the values are snapped to the min and max values after projection.",
    )

    def evaluate(self, spatial_data: anp.ndarray, design_region_dl: float) -> anp.ndarray:
        """Evaluate this transformation on spatial data, given some grid size in the region."""
        filt_proj = make_filter_and_project(
            self.radius, design_region_dl, beta=self.beta, eta=self.eta
        )
        data_projected = filt_proj(spatial_data)

        if self.strict_binarize:
            data_projected = threshold(data_projected)

        return data_projected


TransformationType = typing.Union[FilterProject]



from __future__ import annotations

import typing
from os import PathLike
from pathlib import Path

from tidy3d.components.types.workflow import WorkflowDataType, WorkflowType
from tidy3d.config import config
from tidy3d.log import get_logging_console
from tidy3d.web.api.autograd.autograd import run as run_autograd
from tidy3d.web.api.autograd.autograd import run_async
from tidy3d.web.api.container import DEFAULT_DATA_DIR, DEFAULT_DATA_PATH
from tidy3d.web.core.types import PayType

RunInput: typing.TypeAlias = typing.Union[
    WorkflowType,
    list["RunInput"],
    tuple["RunInput", ...],
    dict[typing.Hashable, "RunInput"],
]

RunOutput: typing.TypeAlias = typing.Union[
    WorkflowDataType,
    list["RunOutput"],
    tuple["RunOutput", ...],
    dict[typing.Hashable, "RunOutput"],
]


def _collect_by_hash(
    node: RunInput,
    found: dict[str, WorkflowType] | None = None,
) -> dict[str, WorkflowType]:
    """Traverses the structure and collects all simulations into a `{hash: sim}` mapping.
    The latest occurrence of the same hash overwrites the previous one — which is fine
    since identical objects share the same hash."""
    if found is None:
        found = {}
    if isinstance(node, WorkflowType):
        found[node._hash_self()] = node
        return found
    if isinstance(node, (list, tuple)):
        for v in node:
            _collect_by_hash(v, found)
        return found
    if isinstance(node, dict):
        if any(isinstance(k, WorkflowType) for k in node.keys()):
            raise ValueError("Dict keys must not be simulations.")
        for v in node.values():
            _collect_by_hash(v, found)
        return found
    raise TypeError(f"Unsupported element in container: {type(node)!r}")


def _reconstruct_by_hash(node: RunInput, h2data: dict[str, WorkflowDataType]) -> RunOutput:
    """Replaces each leaf node (simulation) with its corresponding data object.

    If a simulation appears multiple times in the input structure, the first
    occurrence reuses the same object; subsequent occurrences receive a `.copy()`
    to avoid shared-state side effects.
    """
    seen = set()

    def _recur(n: RunInput) -> RunOutput:
        if isinstance(n, WorkflowType):
            h = n._hash_self()
            data = h2data[h]
            if h in seen:
                return data.copy()
            seen.add(h)
            return data
        if isinstance(n, tuple):
            return tuple(_recur(v) for v in n)
        if isinstance(n, list):
            return [_recur(v) for v in n]
        if isinstance(n, dict):
            return {k: _recur(v) for k, v in n.items()}
        raise TypeError(f"Unsupported element in reconstruction: {type(n)!r}")

    return _recur(node)


def run(
    simulation: RunInput,
    task_name: typing.Optional[str] = None,
    folder_name: str = "default",
    path: typing.Optional[PathLike] = None,
    callback_url: typing.Optional[str] = None,
    verbose: bool = True,
    progress_callback_upload: typing.Optional[typing.Callable[[float], None]] = None,
    progress_callback_download: typing.Optional[typing.Callable[[float], None]] = None,
    solver_version: typing.Optional[str] = None,
    worker_group: typing.Optional[str] = None,
    simulation_type: str = "tidy3d",
    parent_tasks: typing.Optional[list[str]] = None,
    local_gradient: typing.Optional[bool] = None,
    max_num_adjoint_per_fwd: typing.Optional[int] = None,
    reduce_simulation: typing.Literal["auto", True, False] = "auto",
    pay_type: typing.Union[PayType, str] = PayType.AUTO,
    priority: typing.Optional[int] = None,
    max_workers: typing.Optional[int] = None,
    lazy: typing.Optional[bool] = None,
) -> RunOutput:
    """
    Submit one or many simulations and return results in the same container shape.

    This is a convenience wrapper around the autograd runners that accepts a single
    :class:`WorkflowType` **or** an arbitrarily nested container of simulations
    (`list`, `tuple`, or `dict` values). Internally, all simulations are collected,
    deduplicated by object hash, executed either synchronously (single) or
    asynchronously (batch), and the returned data objects are reassembled to mirror
    the input structure.

    **Path behavior**
      - **Single simulation:** results are downloaded to ``f"{path}.hdf5"``.
      - **Multiple simulations:** ``path`` is treated as a **directory**, and each
        task will write its own results file inside that directory.

    **Lazy loading**
      - If ``lazy`` is *not* specified: single runs default to ``False`` (eager load);
        batch runs default to ``True`` (proxy objects that load on first access).

    Parameters
    ----------
    simulation : Union[:class:`.Simulation`, :class:`.HeatSimulation`, :class:`.EMESimulation`] | list | tuple | dict
        A simulation or a container whose leaves are simulations.
        Supported containers are ``list``, ``tuple``, and ``dict`` (values only).
        Dict **keys must not** be simulations.
    task_name : Optional[str], default None
        Optional name for a single run. Prefixed for multiple runs.
    folder_name : str = "default"
        Folder shown on the web UI.
    path : Optional[PathLike] = None
        Output path. Interpreted as a file path for single simulations and a directory for multiple simulations.
        Defaults are "simulation.hdf5" (single simulation) and the current directory (multiple simulations).
    callback_url : Optional[str] = None
        Optional HTTP PUT endpoint to receive completion events.
    verbose : bool = True
        If ``True``, print status and progress; otherwise run quietly.
    progress_callback_upload : Optional[Callable[[float], None]] = None
        Callback invoked with byte counts during upload (single-run path only).
    progress_callback_download : Optional[Callable[[float], None]] = None
        Callback invoked with byte counts during download (single-run path only).
    solver_version : Optional[str] = None
        Target solver version.
    worker_group : Optional[str] = None
        Worker group to target.
    simulation_type : str = "tidy3d"
        Simulation type label passed through to the runners.
    parent_tasks : Optional[List[str]] = None
        Parent task IDs, if any.
    local_gradient : Optional[bool] = None
        Compute gradients locally (more downloads; useful for experimental features).
        Defaults to ``config.adjoint.local_gradient`` when not provided. Remote gradients
        always use server-side defaults.
    max_num_adjoint_per_fwd : Optional[int] = None
        Maximum number of adjoint simulations allowed per forward run. Defaults to
        ``config.adjoint.max_adjoint_per_fwd`` when not provided.
    reduce_simulation : {"auto", True, False} = "auto"
        Whether to reduce structures to the simulation domain (mode solver only).
    pay_type : Union[PayType, str] = PayType.AUTO
        Payment method selection.
    priority : Optional[int] = None
        Queue priority for vGPU (1 = lowest, 10 = highest).
    max_workers : Optional[int] = None
        Maximum parallel submissions for batch runs. ``None`` submits all at once.
    lazy : Optional[bool] = None
        Whether to load the actual data (``lazy=False``) or return a proxy that loads
        the data when accessed (``lazy=True``).

    Returns
    -------
    RunOutput
        A data object (or nested container of data objects) matching the input
        container shape. Leaves are instances of the corresponding
        :class:`WorkflowDataType`.

    Notes
    -----
    - Simulations are indexed by ``hash(sim)``. If the *same object* appears multiple
      times in the input, it is executed once and its data is reused at all positions.
      The *last* occurrence wins if duplicates with the same hash are encountered.
    - For each simulation, a mode-solver compatibility patch is applied so that
      the returned data exposes expected convenience attributes.
    - ``progress_callback_*`` are only used in the single-run code path.

    Raises
    ------
    ValueError
        If no simulations are found in ``simulation``.
    TypeError
        If an unsupported container element is encountered, or if a dict key is a
        simulation object.

    Examples
    --------
        Single run (eager by default)::

        .. code-block:: python

            sim_data = run(sim, task_name="wg_bend", path="out/bend")
            # writes: "out/bend.hdf5"

        Batch run with nested structure (lazy by default)::

        .. code-block:: python

            sims = {
                "coarse": [sim_a, sim_b],
                "fine": sim_c,
            }
            data = run(sims, path="out/batch_dir", max_workers=4)

            # 'data' mirrors 'sims' structure:
            # data["coarse"][0] -> data for sim_a, etc.

    See Also
    --------
    tidy3d.web.api.autograd.autograd.run
        Underlying autograd single-run implementation.
    tidy3d.web.api.autograd.autograd.run_async
        Underlying autograd batch submission implementation.
    """
    h2sim: dict[str, WorkflowType] = _collect_by_hash(simulation)
    if not h2sim:
        raise ValueError("No simulation data found in simulation input.")

    if local_gradient is None:
        local_gradient = bool(config.adjoint.local_gradient)

    if max_num_adjoint_per_fwd is None:
        max_num_adjoint_per_fwd = config.adjoint.max_adjoint_per_fwd

    key_prefix = ""
    if len(h2sim) == 1:
        if path is not None:
            # user may submit the same simulation multiple times and not specify an extension, but dir path
            if not Path(path).suffixes:
                path = f"{path}.hdf5"
                console = get_logging_console()
                console.log(f"Changed output path to {path}")
        else:
            path = DEFAULT_DATA_PATH
        h, sim = next(iter(h2sim.items()))
        data = {
            h: run_autograd(
                simulation=sim,
                task_name=task_name,
                folder_name=folder_name,
                path=path,
                callback_url=callback_url,
                verbose=verbose,
                progress_callback_upload=progress_callback_upload,
                progress_callback_download=progress_callback_download,
                solver_version=solver_version,
                worker_group=worker_group,
                simulation_type=simulation_type,
                parent_tasks=parent_tasks,
                local_gradient=local_gradient,
                max_num_adjoint_per_fwd=max_num_adjoint_per_fwd,
                reduce_simulation=reduce_simulation,
                pay_type=pay_type,
                priority=priority,
                lazy=lazy if lazy is not None else False,
            )
        }
    else:
        key_prefix = f"{task_name}_" if task_name else ""
        sims = {f"{key_prefix}{h}": s for h, s in h2sim.items()}
        path_dir = Path(path) if path is not None else Path(DEFAULT_DATA_DIR)
        data = run_async(
            simulations=sims,
            folder_name=folder_name,
            path_dir=path_dir,
            callback_url=callback_url,
            num_workers=max_workers,
            verbose=verbose,
            simulation_type=simulation_type,
            solver_version=solver_version,
            parent_tasks=parent_tasks,
            local_gradient=local_gradient,
            max_num_adjoint_per_fwd=max_num_adjoint_per_fwd,
            reduce_simulation=reduce_simulation,
            pay_type=pay_type,
            priority=priority,
            lazy=lazy if lazy is not None else True,
        )

    h2data: dict[str, WorkflowDataType] = {}
    for k, sim_data in data.items():
        h = k[len(key_prefix) :] if key_prefix else k
        h2data[h] = sim_data

    return _reconstruct_by_hash(simulation, h2data)



"""Module containing specifications for path integrals."""

from __future__ import annotations

from abc import ABC, abstractmethod

import numpy as np
import pydantic.v1 as pd
import shapely
import xarray as xr
from typing_extensions import Self

from tidy3d.components.base import cached_property
from tidy3d.components.geometry.base import Box, Geometry
from tidy3d.components.microwave.base import MicrowaveBaseModel
from tidy3d.components.types import ArrayFloat2D, Bound, Coordinate, Coordinate2D
from tidy3d.components.types.base import Axis, Direction
from tidy3d.components.validators import assert_line
from tidy3d.constants import MICROMETER, fp_eps
from tidy3d.exceptions import SetupError


class AbstractAxesRH(MicrowaveBaseModel, ABC):
    """Represents an axis-aligned right-handed coordinate system with one axis preferred.
    Typically `main_axis` would refer to the normal axis of a plane.
    """

    @cached_property
    @abstractmethod
    def main_axis(self) -> Axis:
        """Get the preferred axis."""

    @cached_property
    def remaining_axes(self) -> tuple[Axis, Axis]:
        """Get in-plane axes, ordered to maintain a right-handed coordinate system."""
        axes: list[Axis] = [0, 1, 2]
        axes.pop(self.main_axis)
        if self.main_axis == 1:
            return (axes[1], axes[0])
        else:
            return (axes[0], axes[1])

    @cached_property
    def remaining_dims(self) -> tuple[str, str]:
        """Get in-plane dimensions, ordered to maintain a right-handed coordinate system."""
        dim1 = "xyz"[self.remaining_axes[0]]
        dim2 = "xyz"[self.remaining_axes[1]]
        return (dim1, dim2)

    @cached_property
    def local_dims(self) -> tuple[str, str, str]:
        """Get in-plane dimensions with in-plane dims first, followed by the `main_axis` dimension."""
        dim3 = "xyz"[self.main_axis]
        return self.remaining_dims + tuple(dim3)


class AxisAlignedPathIntegralSpec(AbstractAxesRH, Box):
    """Class for defining the simplest type of path integral, which is aligned with Cartesian axes.

    Example
    -------
    >>> path_spec = AxisAlignedPathIntegralSpec(
    ...     center=(0, 0, 1),
    ...     size=(0, 0, 2),
    ...     extrapolate_to_endpoints=True
    ... )
    """

    _line_validator = assert_line()

    extrapolate_to_endpoints: bool = pd.Field(
        False,
        title="Extrapolate to Endpoints",
        description="If the endpoints of the path integral terminate at or near a material interface, "
        "the field is likely discontinuous. When this field is ``True``, fields that are outside and on the bounds "
        "of the integral are ignored. Should be enabled when computing voltage between two conductors.",
    )

    snap_path_to_grid: bool = pd.Field(
        False,
        title="Snap Path to Grid",
        description="It might be desirable to integrate exactly along the Yee grid associated with "
        "a field. When this field is ``True``, the integration path will be snapped to the grid.",
    )

    @cached_property
    def main_axis(self) -> Axis:
        """Axis for performing integration."""
        for index, value in enumerate(self.size):
            if value != 0:
                return index

    def _vertices_2D(self, axis: Axis) -> tuple[Coordinate2D, Coordinate2D]:
        """Returns the two vertices of this path in the plane defined by ``axis``."""
        min = self.bounds[0]
        max = self.bounds[1]
        _, min = Box.pop_axis(min, axis)
        _, max = Box.pop_axis(max, axis)

        u = [min[0], max[0]]
        v = [min[1], max[1]]
        return (u, v)


class Custom2DPathIntegralSpec(AbstractAxesRH):
    """Class for specifying a custom path integral defined as a curve on an axis-aligned plane.

    Example
    -------
    >>> import numpy as np
    >>> vertices = np.array([[0, 0], [1, 0], [1, 1], [0, 1]])
    >>> path_spec = Custom2DPathIntegralSpec(
    ...     axis=2,
    ...     position=0.5,
    ...     vertices=vertices
    ... )

    Notes
    -----

    Given a set of vertices :math:`\\vec{r}_i`, this class approximates path integrals over
    vector fields of the form :math:`\\int{\\vec{F} \\cdot \\vec{dl}}`
    as :math:`\\sum_i{\\vec{F}(\\vec{r}_i) \\cdot \\vec{dl}_i}`,
    where the differential length :math:`\\vec{dl}` is approximated using central differences
    :math:`\\vec{dl}_i = \\frac{\\vec{r}_{i+1} - \\vec{r}_{i-1}}{2}`.
    If the path is not closed, forward and backward differences are used at the endpoints.
    """

    axis: Axis = pd.Field(
        ..., title="Axis", description="Specifies dimension of the planar axis (0,1,2) -> (x,y,z)."
    )

    position: float = pd.Field(
        ...,
        title="Position",
        description="Position of the plane along the ``axis``.",
    )

    vertices: ArrayFloat2D = pd.Field(
        ...,
        title="Vertices",
        description="List of (d1, d2) defining the 2 dimensional positions of the path. "
        "The index of dimension should be in the ascending order, which means "
        "if the axis corresponds with ``y``, the coordinates of the vertices should be (x, z). "
        "If you wish to indicate a closed contour, the final vertex should be made "
        "equal to the first vertex, i.e., ``vertices[-1] == vertices[0]``",
        units=MICROMETER,
    )

    @staticmethod
    def _compute_dl_component(coord_array: xr.DataArray, closed_contour=False) -> np.ndarray:
        """Computes the differential length element along the integration path."""
        dl = np.gradient(coord_array)
        if closed_contour:
            # If the contour is closed, we can use central difference on the starting/end point
            # which will be more accurate than the default forward/backward choice in np.gradient
            grad_end = np.gradient([coord_array[-2], coord_array[0], coord_array[1]])
            dl[0] = dl[-1] = grad_end[1]
        return dl

    @classmethod
    def from_circular_path(
        cls, center: Coordinate, radius: float, num_points: int, normal_axis: Axis, clockwise: bool
    ) -> Self:
        """Creates a ``Custom2DPathIntegralSpec`` from a circular path given a desired number of points
        along the perimeter.

        Parameters
        ----------
        center : Coordinate
            The center of the circle.
        radius : float
            The radius of the circle.
        num_points : int
            The number of equidistant points to use along the perimeter of the circle.
        normal_axis : Axis
            The axis normal to the defined circle.
        clockwise : bool
            When ``True``, the points will be ordered clockwise with respect to the positive
            direction of the ``normal_axis``.

        Returns
        -------
        :class:`.Custom2DPathIntegralSpec`
            A path integral defined on a circular path.
        """

        def generate_circle_coordinates(radius: float, num_points: int, clockwise: bool):
            """Helper for generating x,y vertices around a circle in the local coordinate frame."""
            sign = 1.0
            if clockwise:
                sign = -1.0
            angles = np.linspace(0, sign * 2 * np.pi, num_points, endpoint=True)
            xt = radius * np.cos(angles)
            yt = radius * np.sin(angles)
            return (xt, yt)

        # Get transverse axes
        normal_center, trans_center = Geometry.pop_axis(center, normal_axis)

        # These x,y coordinates in the local coordinate frame
        if normal_axis == 1:
            # Handle special case when y is the axis that is popped
            clockwise = not clockwise
        xt, yt = generate_circle_coordinates(radius, num_points, clockwise)
        xt += trans_center[0]
        yt += trans_center[1]
        circle_vertices = np.column_stack((xt, yt))
        # Close the contour exactly
        circle_vertices[-1, :] = circle_vertices[0, :]
        return cls(axis=normal_axis, position=normal_center, vertices=circle_vertices)

    @cached_property
    def is_closed_contour(self) -> bool:
        """Returns ``true`` when the first vertex equals the last vertex."""
        return np.isclose(
            self.vertices[0, :],
            self.vertices[-1, :],
            rtol=fp_eps,
            atol=np.finfo(np.float32).smallest_normal,
        ).all()

    @cached_property
    def main_axis(self) -> Axis:
        """Axis for performing integration."""
        return self.axis

    @pd.validator("vertices", always=True)
    def _correct_shape(cls, val):
        """Makes sure vertices size is correct."""
        # overall shape of vertices
        if val.shape[1] != 2:
            raise SetupError(
                "'Custom2DPathIntegralSpec.vertices' must be a 2 dimensional array shaped (N, 2). "
                f"Given array with shape of '{val.shape}'."
            )
        return val

    @cached_property
    def bounds(self) -> Bound:
        """Helper to get the geometric bounding box of the path integral."""
        path_min = np.amin(self.vertices, axis=0)
        path_max = np.amax(self.vertices, axis=0)
        min_bound = Geometry.unpop_axis(self.position, path_min, self.axis)
        max_bound = Geometry.unpop_axis(self.position, path_max, self.axis)
        return (min_bound, max_bound)

    @cached_property
    def sign(self) -> Direction:
        """Uses the ordering of the vertices to determine the direction of the current flow."""
        linestr = shapely.LineString(coordinates=self.vertices)
        is_ccw = shapely.is_ccw(linestr)
        # Invert statement when the vertices are given as (x, z)
        if self.axis == 1:
            is_ccw = not is_ccw
        if is_ccw:
            return "+"
        else:
            return "-"



"""Post-processing data and figures of merit for antennas, including radiation efficiency,
reflection efficiency, gain, and realized gain.
"""

from __future__ import annotations

from typing import Literal, Optional

import numpy as np
import pydantic.v1 as pd
import xarray as xr
from typing_extensions import Self

from tidy3d.components.data.data_array import (
    FieldProjectionAngleDataArray,
    FreqDataArray,
    FreqModeDataArray,
    ImpedanceFreqModeDataArray,
)
from tidy3d.components.data.monitor_data import DirectivityData, ModeData, ModeSolverData
from tidy3d.components.microwave.base import MicrowaveBaseModel
from tidy3d.components.microwave.data.data_array import (
    AttenuationConstantArray,
    GroupVelocityArray,
    PhaseConstantArray,
    PhaseVelocityArray,
    PropagationConstantArray,
)
from tidy3d.components.microwave.data.dataset import TransmissionLineDataset
from tidy3d.components.microwave.monitor import MicrowaveModeMonitor, MicrowaveModeSolverMonitor
from tidy3d.components.types import FreqArray, ModeClassification, PolarizationBasis
from tidy3d.constants import C_0
from tidy3d.log import log


class AntennaMetricsData(DirectivityData, MicrowaveBaseModel):
    """Data representing the main parameters and figures of merit for antennas.

    Example
    -------
    >>> import numpy as np
    >>> from tidy3d.components.data.monitor_data import FluxDataArray, FieldProjectionAngleDataArray
    >>> from tidy3d.components.monitor import DirectivityMonitor
    >>> f = np.linspace(1e14, 2e14, 10)
    >>> r = np.atleast_1d(1e6)
    >>> theta = np.linspace(0, np.pi, 10)
    >>> phi = np.linspace(0, 2*np.pi, 20)
    >>> coords = dict(r=r, theta=theta, phi=phi, f=f)
    >>> coords_flux = dict(f=f)
    >>> field_values = (1+1j) * np.random.random((len(r), len(theta), len(phi), len(f)))
    >>> flux_data = FluxDataArray(np.random.random(len(f)), coords=coords_flux)
    >>> scalar_field = FieldProjectionAngleDataArray(field_values, coords=coords)
    >>> monitor = DirectivityMonitor(
    ...     center=(1,2,3),
    ...     size=(2,2,2),
    ...     freqs=f,
    ...     name="rad_monitor",
    ...     phi=phi,
    ...     theta=theta
    ... )
    >>> power_data = FreqDataArray(np.random.random(len(f)), coords=coords_flux)
    >>> data = AntennaMetricsData(
    ...     monitor=monitor,
    ...     projection_surfaces=monitor.projection_surfaces,
    ...     flux=flux_data,
    ...     Er=scalar_field,
    ...     Etheta=scalar_field,
    ...     Ephi=scalar_field,
    ...     Hr=scalar_field,
    ...     Htheta=scalar_field,
    ...     Hphi=scalar_field,
    ...     power_incident=power_data,
    ...     power_reflected=power_data
    ... )

    Notes
    -----
    The definitions of radiation efficiency, reflection efficiency, gain, and realized gain
    are based on:

    Balanis, Constantine A., "Antenna Theory: Analysis and Design,"
    John Wiley & Sons, Chapter 2.9 (2016).
    """

    power_incident: FreqDataArray = pd.Field(
        ...,
        title="Power incident",
        description="Array of values representing the incident power to an antenna.",
    )

    power_reflected: FreqDataArray = pd.Field(
        ...,
        title="Power reflected",
        description="Array of values representing power reflected due to an impedance mismatch with the antenna.",
    )

    @staticmethod
    def from_directivity_data(
        dir_data: DirectivityData, power_inc: FreqDataArray, power_refl: FreqDataArray
    ) -> AntennaMetricsData:
        """Create :class:`.AntennaMetricsData` from directivity data and power measurements.

        Parameters
        ----------
        dir_data : :class:`.DirectivityData`
            Directivity data containing field components and flux measurements.
        power_inc : :class:`.FreqDataArray`
            Array of values representing the incident power to an antenna.
        power_refl : :class:`.FreqDataArray`
            Array of values representing power reflected due to impedance mismatch with the antenna.

        Returns
        -------
        :class:`.AntennaMetricsData`
            New instance combining directivity data with incident and reflected power measurements.
        """
        antenna_params_dict = {
            **dir_data.dict(),
            "power_incident": power_inc,
            "power_reflected": power_refl,
        }
        antenna_params_dict.pop("type")
        return AntennaMetricsData(**antenna_params_dict)

    @property
    def supplied_power(self) -> FreqDataArray:
        """The power supplied to the antenna, which takes into account reflections."""
        return self.power_incident - self.power_reflected

    @property
    def radiation_efficiency(self) -> FreqDataArray:
        """The radiation efficiency of the antenna."""
        return self.calc_radiation_efficiency(self.supplied_power)

    @property
    def reflection_efficiency(self) -> FreqDataArray:
        """The reflection efficiency of the antenna, which is due to an impedance mismatch."""
        reflection_efficiency = self.supplied_power / self.power_incident
        return reflection_efficiency

    def partial_gain(
        self, pol_basis: PolarizationBasis = "linear", tilt_angle: Optional[float] = None
    ) -> xr.Dataset:
        """The partial gain figures of merit for antennas. The partial gains are computed
        in the ``linear`` or ``circular`` polarization bases. If ``tilt_angle`` is not ``None``,
        the partial directivity is computed in the linear polarization basis rotated by ``tilt_angle``
        from the theta-axis. Gain is dimensionless.

        Parameters
        ----------
        pol_basis : PolarizationBasis
            The desired polarization basis used to express partial gain, either
            ``linear`` or ``circular``.

        tilt_angle : float
            The angle by which the co-polar vector is rotated from the theta-axis.
            At ``tilt_angle`` = 0, the co-polar vector coincides with the theta-axis and the cross-polar
            vector coincides with the phi-axis; while at ``tilt_angle = pi/2``, the co-polar vector
            coincides with the phi-axis.

        Returns
        -------
        ``xarray.Dataset``
            Dataset containing the partial gains split into the two polarization states.
        """
        self._check_valid_pol_basis(pol_basis, tilt_angle)
        partial_D = self.partial_directivity(pol_basis=pol_basis, tilt_angle=tilt_angle)
        if pol_basis == "linear":
            if tilt_angle is None:
                rename_mapping = {"Dtheta": "Gtheta", "Dphi": "Gphi"}
            else:
                rename_mapping = {"Dco": "Gco", "Dcross": "Gcross"}
        else:
            rename_mapping = {"Dright": "Gright", "Dleft": "Gleft"}
        return self.radiation_efficiency * partial_D.rename(rename_mapping)

    @property
    def gain(self) -> FieldProjectionAngleDataArray:
        """The gain figure of merit for antennas. Gain is dimensionless."""
        partial_G = self.partial_gain()
        return partial_G.Gtheta + partial_G.Gphi

    def partial_realized_gain(
        self, pol_basis: PolarizationBasis = "linear", tilt_angle: Optional[float] = None
    ) -> xr.Dataset:
        """The partial realized gain figures of merit for antennas. The partial gains are computed
        in the ``linear`` or ``circular`` polarization bases. If ``tilt_angle`` is not ``None``,
        the partial directivity is computed in the linear polarization basis rotated by ``tilt_angle``
        from the theta-axis. Gain is dimensionless.

        Parameters
        ----------
        pol_basis : PolarizationBasis
            The desired polarization basis used to express partial gain, either
            ``linear`` or ``circular``.

        tilt_angle : float
            The angle by which the co-polar vector is rotated from the theta-axis.
            At ``tilt_angle`` = 0, the co-polar vector coincides with the theta-axis and the cross-polar
            vector coincides with the phi-axis; while at ``tilt_angle = pi/2``, the co-polar vector
            coincides with the phi-axis.

        Returns
        -------
        ``xarray.Dataset``
            Dataset containing the partial realized gains split into the two polarization states.
        """
        self._check_valid_pol_basis(pol_basis, tilt_angle)
        reflection_efficiency = self.reflection_efficiency
        partial_G = self.partial_gain(pol_basis=pol_basis, tilt_angle=tilt_angle)
        return reflection_efficiency * partial_G

    @property
    def realized_gain(self) -> FieldProjectionAngleDataArray:
        """The realized gain figure of merit for antennas. Realized gain is dimensionless."""
        partial_G = self.partial_realized_gain()
        return partial_G.Gtheta + partial_G.Gphi


class MicrowaveModeDataBase(MicrowaveBaseModel):
    """Base class for microwave mode data that extends standard mode data with RF/microwave features.

    This base class adds microwave-specific functionality to mode data classes, including:

    - **Transmission line data**: Characteristic impedance (Z0), voltage coefficients, and
      current coefficients for transmission line analysis
    - **Enhanced modes_info**: Includes impedance data in the mode properties dataset
    - **Group index handling**: Properly filters transmission line data when computing group indices
    - **Mode reordering**: Ensures transmission line data tracks with reordered modes

    Notes
    -----
    This is a mixin class that must be combined with mode data classes (:class:`.ModeData` or
    :class:`.ModeSolverData`). It uses ``super()`` to call methods on the mixed-in class, extending
    their functionality rather than replacing it.

    The mixin should be placed first in the inheritance list to ensure its method overrides
    are used.
    """

    transmission_line_data: Optional[TransmissionLineDataset] = pd.Field(
        None,
        title="Transmission Line Data",
        description="Additional data relevant to transmission lines in RF and microwave applications, "
        "like characteristic impedance. This field is populated when a :class:`MicrowaveModeSpec` has "
        "been used to set up the monitor or mode solver.",
    )

    @property
    def modes_info(self) -> xr.Dataset:
        """Dataset collecting various properties of the stored modes."""
        super_info = super().modes_info

        # Add transmission line data if present
        if self.transmission_line_data is not None:
            super_info["Re(Z0)"] = self.transmission_line_data.Z0.real
            super_info["Im(Z0)"] = self.transmission_line_data.Z0.imag
        return super_info

    @property
    def mode_classifications(self) -> list[ModeClassification]:
        """List of mode classifications (TEM, quasi-TEM, TE, TM, or Hybrid) for each mode."""
        return [self._classify_mode(mode_index) for mode_index in self.n_complex.mode_index]

    @property
    def free_space_wavenumber(self) -> FreqDataArray:
        """The free space wavenumber (k_0) in rad/m."""
        freqs = self.n_complex.f.values
        C_0_meters = C_0 * 1e-6
        return FreqDataArray(2 * np.pi * freqs / C_0_meters, coords={"f": freqs})

    @property
    def gamma(self) -> PropagationConstantArray:
        r"""The propagation constant with SI units.

        In the physics convention, where time-harmonic fields evolve with :math:`e^{-j\omega t}`,
        a wave propagating in the +z direction varies as:

        .. math::

           E(z) = E_0 e^{\gamma z} = E_0 e^{-\alpha z} e^{j\beta z}

        where :math:`\gamma = -\alpha + j\beta`.
        """
        data = 1j * self.n_complex * self.free_space_wavenumber
        return PropagationConstantArray(data, coords=self.n_complex.coords)

    @property
    def alpha(self) -> AttenuationConstantArray:
        r"""The attenuation constant (real part of :math:`-\gamma`).

        Causes exponential decay of the field amplitude:

        .. math::

           E(z) = E_0 e^{-\alpha z} e^{j\beta z}

        Units: Nepers/meter (Np/m).
        """
        return -self.gamma.real

    @property
    def beta(self) -> PhaseConstantArray:
        r"""The phase constant (imaginary part of :math:`\gamma`).

        Determines the phase variation of the field:

        .. math::

           E(z) = E_0 e^{-\alpha z} e^{j\beta z}

        Units: radians/meter (rad/m).
        """
        return self.gamma.imag

    @property
    def distance_40dB(self) -> FreqModeDataArray:
        r"""Distance at which the field amplitude drops by 40 dB.

        For a lossy transmission line, this is the distance where the signal
        attenuates by 40 dB:

        .. math::

           d_{40\text{dB}} = \frac{40\,\text{dB}}{20 \log_{10}(e) \cdot \alpha} = \frac{40}{8.686 \cdot \alpha}

        where :math:`\alpha` is the attenuation constant in Nepers/meter.

        Units: meters.
        """
        # Convert attenuation from Nepers/m to dB/m: dB/m = 20*log10(e)*Np/m ≈ 8.686*Np/m
        # Then: distance_40dB = 40 dB / (attenuation in dB/m)
        attenuation_dB_per_m = 20 * np.log10(np.e) * self.alpha
        distance_meters = 40 / attenuation_dB_per_m
        return FreqModeDataArray(distance_meters.values, coords=self.alpha.coords)

    @property
    def effective_relative_permittivity(self) -> FreqModeDataArray:
        """Effective relative permittivity (real part of n_eff²)."""
        e_r_complex = self.n_complex * self.n_complex
        return FreqModeDataArray(e_r_complex.values, coords=self.n_complex.coords)

    @property
    def phase_velocity(self) -> PhaseVelocityArray:
        """Phase velocity (v_p = c/n_eff) in m/s."""
        C_0_meters = C_0 * 1e-6
        v_p = C_0_meters / self.n_eff
        return PhaseVelocityArray(v_p.values, coords=self.n_eff.coords)

    @property
    def group_velocity(self) -> Optional[GroupVelocityArray]:
        """Group velocity (v_g = c/n_group) in m/s."""
        if self.n_group_raw is None:
            log.warning(
                "The 'group_velocity' was not computed. To calculate 'group_velocity' index, pass "
                "'group_index_step = True' in the 'MicrowaveModeSpec'.",
                log_once=True,
            )
            return None
        C_0_meters = C_0 * 1e-6
        v_g = C_0_meters / self.n_group
        return GroupVelocityArray(v_g.values, coords=self.n_eff.coords)

    @property
    def wave_impedance(self) -> ImpedanceFreqModeDataArray:
        r"""Compute the wave impedance associated with the waveguide mode.
        The wave impedance is defined as:

        .. math::

           Z_{\rm wave} = \frac{\int |E_t|^2 \, {\rm d}S}{2 P}.

        where :math:`E_t` is the transverse electric field and :math:`P` is the complex power flow.
        """
        self._check_fields_stored(["Ex", "Ey", "Ez", "Hx", "Hy", "Hz"])

        tan_fields = self._colocated_tangential_fields
        dim1, dim2 = self._tangential_dims
        e1 = tan_fields["E" + dim1]
        e2 = tan_fields["E" + dim2]
        diff_area = self._diff_area
        field_int = [np.abs(e_field) ** 2 for e_field in [e1, e2]]
        tangential_intensity = (diff_area * (field_int[0] + field_int[1])).sum(
            dim=self._tangential_dims
        )
        direction = self.monitor.store_fields_direction
        P = self.complex_flux if direction == "+" else -self.complex_flux
        Z_wave = tangential_intensity / P / 2
        return ImpedanceFreqModeDataArray(Z_wave.values, coords=self.flux.coords)

    def _classify_mode(self, mode_index: int) -> ModeClassification:
        """Classify mode as TEM, quasi-TEM, TE, TM, or Hybrid based on TE/TM fractions."""
        # Make quasi-TEM classification choice based on lowest frequency available
        min_f_idx = self.wg_TE_fraction.f.argmin()
        low_f_TE_frac = self.wg_TE_fraction.sel(mode_index=mode_index).isel(f=min_f_idx).values
        low_f_TM_frac = self.wg_TM_fraction.sel(mode_index=mode_index).isel(f=min_f_idx).values
        # Otherwise we use the average value of the fraction across frequencies
        mean_TE_frac = self.wg_TE_fraction.sel(mode_index=mode_index).mean().values
        mean_TM_frac = self.wg_TM_fraction.sel(mode_index=mode_index).mean().values

        if (
            mean_TE_frac >= self.monitor.mode_spec.tem_polarization_threshold
            and mean_TM_frac >= self.monitor.mode_spec.tem_polarization_threshold
        ):
            return "TEM"
        elif (
            low_f_TE_frac >= self.monitor.mode_spec.qtem_polarization_threshold
            and low_f_TM_frac >= self.monitor.mode_spec.qtem_polarization_threshold
        ):
            return "quasi-TEM"
        elif mean_TE_frac >= self.monitor.mode_spec.tem_polarization_threshold:
            return "TE"
        elif mean_TM_frac >= self.monitor.mode_spec.tem_polarization_threshold:
            return "TM"
        else:
            return "Hybrid"

    def _group_index_post_process(self, frequency_step: float) -> Self:
        """Calculate group index and remove added frequencies used only for this calculation.

        Parameters
        ----------
        frequency_step: float
            Fractional frequency step used to calculate the group index.

        Returns
        -------
        Self
            Filtered data with calculated group index.
        """
        super_data = super()._group_index_post_process(frequency_step)

        # Add transmission line data handling if present
        if self.transmission_line_data is not None:
            _, center_inds, _ = self._group_index_freq_slices()
            update_dict = {
                "Z0": self.transmission_line_data.Z0.isel(f=center_inds),
                "voltage_coeffs": self.transmission_line_data.voltage_coeffs.isel(f=center_inds),
                "current_coeffs": self.transmission_line_data.current_coeffs.isel(f=center_inds),
            }
            super_data = super_data.updated_copy(**update_dict, path="transmission_line_data")
        return super_data

    def _apply_mode_reorder(self, sort_inds_2d):
        """Apply a mode reordering along mode_index for all frequency indices.

        Parameters
        ----------
        sort_inds_2d : np.ndarray
            Array of shape (num_freqs, num_modes) where each row is the
            permutation to apply to the mode_index for that frequency.
        """
        main_data_reordered = super()._apply_mode_reorder(sort_inds_2d)

        # Add transmission line data handling if present
        if self.transmission_line_data is not None:
            transmission_line_data_reordered = self.transmission_line_data._apply_mode_reorder(
                sort_inds_2d
            )
            main_data_reordered = main_data_reordered.updated_copy(
                transmission_line_data=transmission_line_data_reordered
            )
        return main_data_reordered


class MicrowaveModeData(MicrowaveModeDataBase, ModeData):
    """
    Data associated with a :class:`.ModeMonitor` for microwave and RF applications: modal amplitudes,
    propagation indices, mode profiles, and transmission line data.

    Notes
    -----

        This class extends :class:`.ModeData` with additional microwave-specific data including
        characteristic impedance, voltage coefficients, and current coefficients. The data is
        stored as `DataArray <https://docs.xarray.dev/en/stable/generated/xarray.DataArray.html>`_
        objects using the `xarray <https://docs.xarray.dev/en/stable/index.html>`_ package.

        The microwave mode data contains all the information from :class:`.ModeData` plus additional
        microwave dataset with impedance calculations performed using voltage and current line integrals
        as specified in the :class:`.MicrowaveModeSpec`.

    Example
    -------
    >>> import tidy3d as td
    >>> import numpy as np
    >>> from tidy3d.components.data.data_array import (
    ...     CurrentFreqModeDataArray,
    ...     ImpedanceFreqModeDataArray,
    ...     ModeAmpsDataArray,
    ...     ModeIndexDataArray,
    ...     VoltageFreqModeDataArray,
    ... )
    >>> from tidy3d.components.microwave.data.dataset import TransmissionLineDataset
    >>> direction = ["+", "-"]
    >>> f = [1e14, 2e14, 3e14]
    >>> mode_index = np.arange(3)
    >>> index_coords = dict(f=f, mode_index=mode_index)
    >>> index_data = ModeIndexDataArray((1+1j) * np.random.random((3, 3)), coords=index_coords)
    >>> amp_coords = dict(direction=direction, f=f, mode_index=mode_index)
    >>> amp_data = ModeAmpsDataArray((1+1j) * np.random.random((2, 3, 3)), coords=amp_coords)
    >>> impedance_data = ImpedanceFreqModeDataArray(50 * np.ones((3, 3)), coords=index_coords)
    >>> voltage_data = VoltageFreqModeDataArray((1+1j) * np.random.random((3, 3)), coords=index_coords)
    >>> current_data = CurrentFreqModeDataArray((0.02+0.01j) * np.random.random((3, 3)), coords=index_coords)
    >>> tl_data = TransmissionLineDataset(
    ...     Z0=impedance_data,
    ...     voltage_coeffs=voltage_data,
    ...     current_coeffs=current_data
    ... )
    >>> monitor = td.MicrowaveModeMonitor(
    ...    center=(0, 0, 0),
    ...    size=(2, 0, 6),
    ...    freqs=[2e14, 3e14],
    ...    mode_spec=td.MicrowaveModeSpec(num_modes=3, impedance_specs=td.AutoImpedanceSpec()),
    ...    name='microwave_mode',
    ... )
    >>> data = MicrowaveModeData(
    ...     monitor=monitor,
    ...     amps=amp_data,
    ...     n_complex=index_data,
    ...     transmission_line_data=tl_data
    ... )
    """

    monitor: MicrowaveModeMonitor = pd.Field(
        ..., title="Monitor", description="Mode monitor associated with the data."
    )


class MicrowaveModeSolverData(MicrowaveModeDataBase, ModeSolverData):
    """
    Data associated with a :class:`.ModeSolverMonitor` for microwave and RF applications: scalar components
    of E and H fields plus characteristic impedance data.

    Notes
    -----

        This class extends :class:`.ModeSolverData` with additional microwave-specific data including
        characteristic impedance, voltage coefficients, and current coefficients. The data is
        stored as `DataArray <https://docs.xarray.dev/en/stable/generated/xarray.DataArray.html>`_
        objects using the `xarray <https://docs.xarray.dev/en/stable/index.html>`_ package.

        The microwave mode solver data contains all field components (Ex, Ey, Ez, Hx, Hy, Hz) and
        effective indices from :class:`.ModeSolverData`, plus impedance calculations performed using
        voltage and current line integrals as specified in the :class:`.MicrowaveModeSpec`.

    Example
    -------
    >>> import tidy3d as td
    >>> import numpy as np
    >>> from tidy3d import Grid, Coords
    >>> from tidy3d.components.data.data_array import (
    ...     CurrentFreqModeDataArray,
    ...     ImpedanceFreqModeDataArray,
    ...     ScalarModeFieldDataArray,
    ...     ModeIndexDataArray,
    ...     VoltageFreqModeDataArray,
    ... )
    >>> from tidy3d.components.microwave.data.dataset import TransmissionLineDataset
    >>> x = [-1, 1, 3]
    >>> y = [-2, 0]
    >>> z = [-3, -1, 1, 3, 5]
    >>> f = [2e14, 3e14]
    >>> mode_index = np.arange(3)
    >>> grid = Grid(boundaries=Coords(x=x, y=y, z=z))
    >>> field_coords = dict(x=x[:-1], y=y[:-1], z=z[:-1], f=f, mode_index=mode_index)
    >>> field = ScalarModeFieldDataArray((1+1j)*np.random.random((2,1,4,2,3)), coords=field_coords)
    >>> index_coords = dict(f=f, mode_index=mode_index)
    >>> index_data = ModeIndexDataArray((1+1j) * np.random.random((2,3)), coords=index_coords)
    >>> impedance_data = ImpedanceFreqModeDataArray(50 * np.ones((2, 3)), coords=index_coords)
    >>> voltage_data = VoltageFreqModeDataArray((1+1j) * np.random.random((2, 3)), coords=index_coords)
    >>> current_data = CurrentFreqModeDataArray((0.02+0.01j) * np.random.random((2, 3)), coords=index_coords)
    >>> tl_data = TransmissionLineDataset(
    ...     Z0=impedance_data,
    ...     voltage_coeffs=voltage_data,
    ...     current_coeffs=current_data
    ... )
    >>> monitor = td.MicrowaveModeSolverMonitor(
    ...    center=(0, 0, 0),
    ...    size=(2, 0, 6),
    ...    freqs=[2e14, 3e14],
    ...    mode_spec=td.MicrowaveModeSpec(num_modes=3, impedance_specs=td.AutoImpedanceSpec()),
    ...    name='microwave_mode_solver',
    ... )
    >>> data = MicrowaveModeSolverData(
    ...     monitor=monitor,
    ...     Ex=field,
    ...     Ey=field,
    ...     Ez=field,
    ...     Hx=field,
    ...     Hy=field,
    ...     Hz=field,
    ...     n_complex=index_data,
    ...     grid_expanded=grid,
    ...     transmission_line_data=tl_data
    ... )
    """

    monitor: MicrowaveModeSolverMonitor = pd.Field(
        ..., title="Monitor", description="Mode monitor associated with the data."
    )

    def interp_in_freq(
        self,
        freqs: FreqArray,
        method: Literal["linear", "cubic", "poly"] = "linear",
        renormalize: bool = True,
        recalculate_grid_correction: bool = True,
        assume_sorted: bool = False,
    ) -> MicrowaveModeData:
        """Interpolate mode data to new frequency points.

        Interpolates all stored mode data (effective indices, field components, group indices,
        and dispersion) from the current frequency grid to a new set of frequencies. This is
        useful for obtaining mode data at many frequencies from computations at fewer frequencies,
        when modes vary smoothly with frequency.

        Parameters
        ----------
        freqs : FreqArray
            New frequency points to interpolate to. Should generally span a similar range
            as the original frequencies to avoid extrapolation.
        method : Literal["linear", "cubic", "cheb"]
            Interpolation method. ``"linear"`` for linear interpolation (requires 2+ source
            frequencies), ``"cubic"`` for cubic spline interpolation (requires 4+ source
            frequencies), ``"cheb"`` for Chebyshev polynomial interpolation using barycentric
            formula (requires 3+ source frequencies at Chebyshev nodes).
            For complex-valued data, real and imaginary parts are interpolated independently.
        renormalize : Optional[bool] = True
            Whether to renormalize the mode profiles to unity power after interpolation.
        recalculate_grid_correction : bool = True
            Whether to recalculate the grid correction factors after interpolation or use interpolated
            grid corrections.
        assume_sorted: bool = False,
            Whether to assume the frequency points are sorted.

        Returns
        -------
        ModeSolverData
            New :class:`ModeSolverData` object with data interpolated to the requested frequencies.

        Raises
        ------
        DataError
            If interpolation parameters are invalid (e.g., too few source frequencies for the
            chosen method, or source frequencies not at Chebyshev nodes for 'cheb' method).

        Note
        ----
            Interpolation assumes modes vary smoothly with frequency. Results may be inaccurate
            near mode crossings or regions of rapid mode variation. Use frequency tracking
            (``mode_spec.sort_spec.track_freq``) to help maintain mode ordering consistency.

            For Chebyshev interpolation, source frequencies must be at Chebyshev nodes of the
            second kind within the frequency range.

        Example
        -------
        >>> # Compute modes at 5 frequencies
        >>> import numpy as np
        >>> freqs_sparse = np.linspace(1e14, 2e14, 5)
        >>> # ... create mode_solver and compute modes ...
        >>> # mode_data = mode_solver.solve()
        >>> # Interpolate to 50 frequencies
        >>> freqs_dense = np.linspace(1e14, 2e14, 50)
        >>> # mode_data_interp = mode_data.interp(freqs=freqs_dense, method='linear')
        """
        main_data_interp = super().interp_in_freq(
            freqs, method, renormalize, recalculate_grid_correction, assume_sorted
        )
        if self.transmission_line_data is not None:
            update_dict = self.transmission_line_data._interp_in_freq_update_dict(
                freqs, method, assume_sorted
            )
            transmission_line_data_interp = self.transmission_line_data.updated_copy(**update_dict)
            main_data_interp = main_data_interp.updated_copy(
                transmission_line_data=transmission_line_data_interp
            )
        return main_data_interp



"""Base class for generating an S matrix automatically from tidy3d simulations and port definitions."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Literal, Optional, Union

import pydantic.v1 as pd

from tidy3d.components.base import Tidy3dBaseModel, cached_property, skip_if_fields_missing
from tidy3d.components.geometry.utils import _shift_value_signed
from tidy3d.components.simulation import Simulation
from tidy3d.components.source.time import SourceTimeType
from tidy3d.components.types import Complex, FreqArray
from tidy3d.components.validators import (
    assert_unique_names,
    validate_freqs_min,
    validate_freqs_not_empty,
    validate_freqs_unique,
)
from tidy3d.config import config
from tidy3d.constants import HERTZ
from tidy3d.exceptions import SetupError, Tidy3dKeyError
from tidy3d.log import log
from tidy3d.plugins.smatrix.ports.modal import Port
from tidy3d.plugins.smatrix.ports.types import LumpedPortType, PortType, TerminalPortType
from tidy3d.plugins.smatrix.ports.wave import WavePort
from tidy3d.plugins.smatrix.types import Element, MatrixIndex, NetworkElement, NetworkIndex

if TYPE_CHECKING:
    from tidy3d.web.core.types import PayType

# fwidth of gaussian pulse in units of central frequency
FWIDTH_FRAC = 1.0 / 10
DEFAULT_DATA_DIR = "."

IndexType = Union[MatrixIndex, NetworkIndex]
ElementType = Union[Element, NetworkElement]
TaskNameFormat = Literal["RF", "PF"]


class AbstractComponentModeler(ABC, Tidy3dBaseModel):
    """Tool for modeling devices and computing port parameters."""

    name: str = pd.Field(
        "",
        title="Name",
    )

    simulation: Simulation = pd.Field(
        ...,
        title="Simulation",
        description="Simulation describing the device without any sources present.",
    )

    ports: tuple[Union[Port, TerminalPortType], ...] = pd.Field(
        (),
        title="Ports",
        description="Collection of ports describing the scattering matrix elements. "
        "For each input mode, one simulation will be run with a modal source.",
    )

    freqs: FreqArray = pd.Field(
        ...,
        title="Frequencies",
        description="Array or list of frequencies at which to compute port parameters.",
        units=HERTZ,
    )

    remove_dc_component: bool = pd.Field(
        True,
        title="Remove DC Component",
        description="Whether to remove the DC component in the Gaussian pulse spectrum. "
        "If ``True``, the Gaussian pulse is modified at low frequencies to zero out the "
        "DC component, which is usually desirable so that the fields will decay. However, "
        "for broadband simulations, it may be better to have non-vanishing source power "
        "near zero frequency. Setting this to ``False`` results in an unmodified Gaussian "
        "pulse spectrum which can have a nonzero DC component.",
    )

    run_only: Optional[tuple[IndexType, ...]] = pd.Field(
        None,
        title="Run Only",
        description="Set of matrix indices that define the simulations to run. "
        "If ``None``, simulations will be run for all indices in the scattering matrix. "
        "If a tuple is given, simulations will be run only for the given matrix indices.",
    )

    element_mappings: tuple[tuple[ElementType, ElementType, Complex], ...] = pd.Field(
        (),
        title="Element Mappings",
        description="Tuple of S matrix element mappings, each described by a tuple of "
        "(input_element, output_element, coefficient), where the coefficient is the "
        "element_mapping coefficient describing the relationship between the input and output "
        "matrix element. If all elements of a given column of the scattering matrix are defined "
        "by ``element_mappings``, the simulation corresponding to this column is skipped automatically.",
    )
    custom_source_time: Optional[SourceTimeType] = pd.Field(
        None,
        title="Custom Source Time",
        description="If provided, this will be used as specification of the source time-dependence in simulations. "
        "Otherwise, a default source time will be constructed.",
    )

    @pd.root_validator(pre=False)
    def _warn_refactor_2_10(cls, values):
        log.warning(
            f"'{cls.__name__}' was refactored (tidy3d 'v2.10.0'). Existing functionality is available differently. Please consult the migration documentation: https://docs.flexcompute.com/projects/tidy3d/en/latest/api/microwave/microwave_migration.html",
            log_once=True,
        )
        return values

    @pd.validator("simulation", always=True)
    def _sim_has_no_sources(cls, val):
        """Make sure simulation has no sources as they interfere with tool."""
        if len(val.sources) > 0:
            raise SetupError(f"'{cls.__name__}.simulation' must not have any sources.")
        return val

    @pd.validator("element_mappings", always=True)
    def _validate_element_mappings(cls, element_mappings, values):
        """
        Validate that each source index referenced in element_mappings is included in run_only.
        """
        run_only = values.get("run_only")
        if run_only is None:
            return element_mappings

        valid_set = set(run_only)
        invalid_indices = set()
        for mapping in element_mappings:
            input_element = mapping[0]
            output_element = mapping[1]
            for source_index in [input_element[1], output_element[1]]:
                if source_index not in valid_set:
                    invalid_indices.add(source_index)
        if invalid_indices:
            raise SetupError(
                f"'element_mappings' references source index(es) {invalid_indices} "
                f"that are not present in run_only: {run_only}."
            )
        return element_mappings

    @pd.validator("run_only", always=True)
    @skip_if_fields_missing(["ports"])
    def _validate_run_only(cls, val, values):
        """Validate that run_only entries are unique and exist in matrix_indices_monitor."""
        if val is None:
            return val

        # Check uniqueness
        if len(val) != len(set(val)):
            duplicates = [idx for idx in set(val) if val.count(idx) > 1]
            raise SetupError(
                f"'run_only' contains duplicate entries: {duplicates}. "
                "Each index must appear only once."
            )

        # Check membership - use the helper method to get valid indices
        ports = values["ports"]

        valid_indices = set(cls._construct_matrix_indices_monitor(ports))
        invalid_indices = [idx for idx in val if idx not in valid_indices]

        if invalid_indices:
            raise SetupError(
                f"'run_only' contains indices {invalid_indices} that are not present in "
                f"'matrix_indices_monitor'. Valid indices are: {sorted(valid_indices)}"
            )

        return val

    _freqs_not_empty = validate_freqs_not_empty()
    _freqs_lower_bound = validate_freqs_min()
    _freqs_unique = validate_freqs_unique()

    @pd.validator("custom_source_time", always=True)
    @skip_if_fields_missing(["freqs"])
    def _freqs_in_custom_source_time(cls, val, values):
        """Make sure freqs is in the range of the custom source time."""
        if val is None:
            return val
        freq_range = val._frequency_range_sigma_cached
        freqs = values["freqs"]

        if freq_range[0] > min(freqs) or max(freqs) > freq_range[1]:
            log.warning(
                "Custom source time does not cover all 'freqs'.",
            )
        return val

    @staticmethod
    def get_task_name(port: PortType, mode_index: Optional[int] = None) -> str:
        """Generates a standardized task name from a port object.

        This method creates a unique string identifier for a simulation task based on
        a port and, if applicable, a specified mode index.

        Parameters
        ----------
        port : PortType
            The port object from which to derive the base name.
        mode_index : Optional[int], optional
            If provided, this index is appended
            to the port name (e.g., 'port_1@1'). Defaults to `None`, in which case the first
            mode is chosen by default.

        Returns
        -------
        str
            The formatted task name string.

        Raises
        ------
        ValueError
            If `mode_index` is specified for a lumped port.
        """

        if isinstance(port, LumpedPortType):
            if mode_index is not None:
                raise ValueError(
                    "'mode_index' should not be specified for a lumped port, "
                    f"but was passed with value '{mode_index}'."
                )
            return f"{port.name}"
        elif isinstance(port, WavePort):
            # WavePorts default to first mode index
            if mode_index is not None:
                return f"{port.name}@{mode_index}"
            return f"{port.name}@{port._mode_indices[0]}"
        else:
            # Modal ports default to 0
            if mode_index is not None:
                return f"{port.name}@{mode_index}"
            return f"{port.name}@0"

    def get_port_by_name(self, port_name: str) -> Port:
        """Get the port from the name."""
        ports = [port for port in self.ports if port.name == port_name]
        if len(ports) == 0:
            raise Tidy3dKeyError(f'Port "{port_name}" not found.')
        return ports[0]

    @staticmethod
    @abstractmethod
    def _construct_matrix_indices_monitor(ports: tuple) -> tuple[IndexType, ...]:
        """Construct matrix indices for monitoring from ports.

        This helper method is used by both the matrix_indices_monitor property
        and the run_only validator to ensure consistency.

        Parameters
        ----------
        ports : tuple
            Tuple of port objects.

        Returns
        -------
        tuple[IndexType, ...]
            Tuple of matrix indices for monitoring.
        """

    @property
    @abstractmethod
    def matrix_indices_monitor(self) -> tuple[IndexType, ...]:
        """Abstract property for all matrix indices that will be used to collect data."""

    @cached_property
    def matrix_indices_source(self) -> tuple[IndexType, ...]:
        """Tuple of all the source matrix indices, which may be less than the total number of
        ports."""
        if self.run_only is not None:
            return self.run_only
        return self.matrix_indices_monitor

    @cached_property
    def matrix_indices_run_sim(self) -> tuple[IndexType, ...]:
        """Tuple of all the matrix indices that will be used to run simulations."""

        if not self.element_mappings:
            return self.matrix_indices_source

        # all the (i, j) pairs in `S_ij` that are tagged as covered by `element_mappings`
        elements_determined_by_map = [element_out for (_, element_out, _) in self.element_mappings]

        # loop through rows of the full s matrix and record rows that still need running.
        source_indices_needed = []
        for col_index in self.matrix_indices_source:
            # loop through columns and keep track of whether each element is covered by mapping.
            matrix_elements_covered = []
            for row_index in self.matrix_indices_monitor:
                element = (row_index, col_index)
                element_covered_by_map = element in elements_determined_by_map
                matrix_elements_covered.append(element_covered_by_map)

            # if any matrix elements in row still not covered by map, a source is needed for row.
            if not all(matrix_elements_covered):
                source_indices_needed.append(col_index)

        return source_indices_needed

    def _shift_value_signed(self, port: Union[Port, WavePort]) -> float:
        """How far (signed) to shift the source from the monitor."""

        return _shift_value_signed(
            obj=port,
            grid=self.simulation.grid,
            bounds=self.simulation.bounds,
            direction=port.direction,
            shift=-2,
            name=f"Port {port.name}",
        )

    unique_port_names = assert_unique_names("ports")

    def run(
        self,
        path_dir: str = DEFAULT_DATA_DIR,
        *,
        folder_name: str = "default",
        callback_url: Optional[str] = None,
        verbose: bool = True,
        solver_version: Optional[str] = None,
        pay_type: Union[PayType, str] = "AUTO",
        priority: Optional[int] = None,
        local_gradient: bool = False,
        max_num_adjoint_per_fwd: Optional[int] = None,
    ):
        log.warning(
            "'ComponentModeler.run()' is deprecated and will be removed in a future release. "
            "Use web.run(modeler) instead. 'web.run' returns a 'ComponentModelerData' object; "
            "get the scattering matrix via 'data.smatrix()'.",
            log_once=True,
        )
        from tidy3d.plugins.smatrix.run import _run_local

        if max_num_adjoint_per_fwd is None:
            max_num_adjoint_per_fwd = config.adjoint.max_adjoint_per_fwd

        data = _run_local(
            self,
            path_dir=path_dir,
            folder_name=folder_name,
            callback_url=callback_url,
            verbose=verbose,
            solver_version=solver_version,
            pay_type=pay_type,
            priority=priority,
            local_gradient=local_gradient,
            max_num_adjoint_per_fwd=max_num_adjoint_per_fwd,
        )
        return data.smatrix()

    def validate_pre_upload(self):
        """Validate the modeler before upload."""
        self.base_sim.validate_pre_upload(source_required=False)


AbstractComponentModeler.update_forward_refs()



from __future__ import annotations

from abc import ABC, abstractmethod

import numpy as np
import pydantic.v1 as pd

from tidy3d.components.base import Tidy3dBaseModel
from tidy3d.constants import HBAR, K_B, M_E_EV, PERCMCUBE
from tidy3d.exceptions import DataError

um_3_to_cm_3 = 1e12  # conversion factor from micron^(-3) to cm^(-3)
DOS_aux_const = 2.0 * np.power((M_E_EV * K_B) / (2 * np.pi * HBAR * HBAR), 1.5) * um_3_to_cm_3


class EffectiveDOS(Tidy3dBaseModel, ABC):
    """Abstract class for the effective density of states"""

    @abstractmethod
    def calc_eff_dos(self, T: float) -> None:
        """Abstract method to calculate the effective density of states."""

    def get_effective_DOS(self, T: float):
        if T <= 0:
            raise DataError(
                f"Incorrect temperature value ({T}) for the effective density of states calculation."
            )

        return self.calc_eff_dos(T)


class ConstantEffectiveDOS(EffectiveDOS):
    """Constant effective density of states model."""

    N: pd.PositiveFloat = pd.Field(
        ..., title="Effective DOS", description="Effective density of states", units=PERCMCUBE
    )

    def calc_eff_dos(self, T: float):
        return self.N


class IsotropicEffectiveDOS(EffectiveDOS):
    """Effective density of states model that assumes single valley and isotropic effective mass.
    The model assumes the standard equation for the 3D semiconductor with parabolic energy dispersion:

    Notes
    -----

    .. math::

        \\mathbf{N_eff} = 2 * (\\frac{m_eff * m_e * k_B T}{2 \\pi \\hbar^2})^(3/2)

    """

    m_eff: pd.PositiveFloat = pd.Field(
        ...,
        title="Effective mass",
        description="Effective mass of the carriers relative to the electron mass at rest",
    )

    def calc_eff_dos(self, T: float):
        return np.power(self.m_eff * T, 1.5) * DOS_aux_const


class MultiValleyEffectiveDOS(EffectiveDOS):
    """Effective density of states model that assumes multiple equivalent valleys and anisotropic effective mass.
    The model assumes the standard equation for the 3D semiconductor with parabolic energy dispersion:

    Notes
    -----

    .. math::

       N_{\\text{eff}} = 2 N_{\\text{valley}} \\left( m_{\\text{eff,long}} m_{\\text{eff,trans}}^2 \\right)^{1/2} \\left( \\frac{m_e k_B T}{2 \\pi \\hbar^2} \\right)^{3/2}

    """

    m_eff_long: pd.PositiveFloat = pd.Field(
        ...,
        title="Longitudinal effective mass",
        description="Relative effective mass of the carriers in the longitudinal direction. This is a relative value compared to the electron mass at rest.",
    )

    m_eff_trans: pd.PositiveFloat = pd.Field(
        ...,
        title="Transverse effective mass",
        description="Relative effective mass of the carriers in the transverse direction. This is a relative value compared to the electron mass at rest.",
    )

    N_valley: pd.PositiveFloat = pd.Field(
        ..., title="Number of valleys", description="Number of effective valleys"
    )

    def calc_eff_dos(self, T: float):
        return (
            self.N_valley
            * np.power(self.m_eff_long * self.m_eff_trans * self.m_eff_trans, 0.5)
            * np.power(T, 1.5)
            * DOS_aux_const
        )


class DualValleyEffectiveDOS(EffectiveDOS):
    """Effective density of states model that assumes combination of light holes and heavy holes with isotropic effective masses.
    The model assumes the standard equation for the 3D semiconductor with parabolic energy dispersion:

    Notes
    -----

    .. math::

       N_{eff} = 2 \\left( \\frac{m_{\\text{eff, lh}} m_e k_B T}{2 \\pi \\hbar^2} \\right)^{3/2} + 2 \\left( \\frac{m_{\\text{eff, hh}} m_e k_B T}{2 \\pi \\hbar^2} \\right)^{3/2}

    """

    m_eff_lh: pd.PositiveFloat = pd.Field(
        ...,
        title="Light hole effective mass",
        description="Relative effective mass of the light holes. This is a relative value compared to the electron mass at rest.",
    )

    m_eff_hh: pd.PositiveFloat = pd.Field(
        ...,
        title="Heavy hole effective mass",
        description="Relative effective mass of the heavy holes. This is a relative value compared to the electron mass at rest.",
    )

    def calc_eff_dos(self, T: float):
        return (np.power(self.m_eff_lh * T, 1.5) + np.power(self.m_eff_hh * T, 1.5)) * DOS_aux_const



"""Objects that define how data is recorded from simulation."""

from __future__ import annotations

import pydantic.v1 as pydantic

from tidy3d.components.microwave.base import MicrowaveBaseModel
from tidy3d.components.microwave.mode_spec import MicrowaveModeSpec
from tidy3d.components.monitor import ModeMonitor, ModeSolverMonitor


class MicrowaveModeMonitorBase(MicrowaveBaseModel):
    """Base class for microwave mode monitors that use :class:`.MicrowaveModeSpec`.

    This mixin provides the ``mode_spec`` field configured for RF and microwave applications,
    including characteristic impedance calculations and transmission line analysis.

    Notes
    -----
    This is a mixin class that provides the :class:`.MicrowaveModeSpec` field for mode monitors.
    It must be placed first in the inheritance list to ensure its ``mode_spec`` field takes
    precedence over the base :class:`.ModeSpec` field from :class:`.AbstractModeMonitor`.
    """

    mode_spec: MicrowaveModeSpec = pydantic.Field(
        default_factory=MicrowaveModeSpec._default_without_license_warning,
        title="Mode Specification",
        description="Parameters to feed to mode solver which determine modes measured by monitor.",
    )


class MicrowaveModeMonitor(MicrowaveModeMonitorBase, ModeMonitor):
    """:class:`Monitor` that records amplitudes from modal decomposition of fields on plane.

    Notes
    ------

        The fields recorded by frequency monitors (and hence also mode monitors) are automatically
        normalized by the power amplitude spectrum of the source. For multiple sources, the user can
        select which source to use for the normalization too.

        We can also use the mode amplitudes recorded in the mode monitor to reveal the decomposition
        of the radiated power into forward- and backward-propagating modes, respectively.

        .. TODO give an example of how to extract the data from this mode.

        .. TODO add derivation in the notebook.

        .. TODO add link to method

        .. TODO add links to notebooks correspondingly

    Example
    -------
    >>> mode_spec = MicrowaveModeSpec(num_modes=3)
    >>> monitor = MicrowaveModeMonitor(
    ...     center=(1,2,3),
    ...     size=(2,2,0),
    ...     freqs=[200e12, 210e12],
    ...     mode_spec=mode_spec,
    ...     name='mode_monitor')

    See Also
    --------

    **Notebooks**:
        * `ModalSourcesMonitors <../../notebooks/ModalSourcesMonitors.html>`_
    """


class MicrowaveModeSolverMonitor(MicrowaveModeMonitorBase, ModeSolverMonitor):
    """:class:`Monitor` that stores the mode field profiles returned by the mode solver in the
    monitor plane.

    Example
    -------
    >>> mode_spec = MicrowaveModeSpec(num_modes=3)
    >>> monitor = MicrowaveModeSolverMonitor(
    ...     center=(1,2,3),
    ...     size=(2,2,0),
    ...     freqs=[200e12, 210e12],
    ...     mode_spec=mode_spec,
    ...     name='mode_monitor')
    """



from __future__ import annotations

# keys for data into auxiliary dictionary (re-exported in autograd.py for tests)
AUX_KEY_SIM_DATA_ORIGINAL = "sim_data"
AUX_KEY_SIM_DATA_FWD = "sim_data_fwd_adjoint"
AUX_KEY_FWD_TASK_ID = "task_id_fwd"
AUX_KEY_SIM_ORIGINAL = "sim_original"

# server-side auxiliary files to upload/download
SIM_VJP_FILE = "output/autograd_sim_vjp.hdf5"
SIM_FIELDS_KEYS_FILE = "autograd_sim_fields_keys.hdf5"



"""Tidy3D configuration system public API."""

from __future__ import annotations

from typing import Any

from . import sections  # noqa: F401 - ensure builtin sections register
from .legacy import LegacyConfigWrapper, LegacyEnvironment, LegacyEnvironmentConfig
from .manager import ConfigManager
from .registry import (
    get_handlers,
    get_sections,
    register_handler,
    register_plugin,
    register_section,
)

__all__ = [
    "ConfigManager",
    "Env",
    "Environment",
    "EnvironmentConfig",
    "config",
    "get_handlers",
    "get_sections",
    "register_handler",
    "register_plugin",
    "register_section",
]


def _create_manager() -> ConfigManager:
    return ConfigManager()


_base_manager = _create_manager()
# TODO(FXC-3827): Drop LegacyConfigWrapper once legacy accessors are removed in Tidy3D 2.12.
_config_wrapper = LegacyConfigWrapper(_base_manager)
config = _config_wrapper

# TODO(FXC-3827): Remove legacy Env exports after deprecation window (planned 2.12).
Environment = LegacyEnvironment
EnvironmentConfig = LegacyEnvironmentConfig
Env = LegacyEnvironment(_base_manager)


def reload_config(*, profile: str | None = None) -> LegacyConfigWrapper:
    """Recreate the global configuration manager (primarily for tests)."""

    global _base_manager, Env
    if _base_manager is not None:
        try:
            _base_manager.apply_web_env({})
        except AttributeError:
            pass
    _base_manager = ConfigManager(profile=profile)
    _config_wrapper.reset_manager(_base_manager)
    Env.reset_manager(_base_manager)
    return _config_wrapper


def get_manager() -> ConfigManager:
    """Return the underlying configuration manager instance."""

    return _base_manager


def __getattr__(name: str) -> Any:
    return getattr(config, name)



"""Module containing specifications for current path integrals."""

from __future__ import annotations

from typing import Any, Literal, Optional, Union

import numpy as np
import pydantic.v1 as pd

from tidy3d.components.base import cached_property
from tidy3d.components.geometry.base import Box, Geometry
from tidy3d.components.geometry.bound_ops import bounds_union
from tidy3d.components.microwave.base import MicrowaveBaseModel
from tidy3d.components.microwave.path_integrals.specs.base import (
    AbstractAxesRH,
    AxisAlignedPathIntegralSpec,
    Custom2DPathIntegralSpec,
)
from tidy3d.components.microwave.path_integrals.viz import ARROW_CURRENT, plot_params_current_path
from tidy3d.components.types import Ax, Bound
from tidy3d.components.types.base import Axis, Direction
from tidy3d.components.validators import assert_plane
from tidy3d.components.viz import add_ax_if_none
from tidy3d.constants import fp_eps
from tidy3d.exceptions import SetupError


class AxisAlignedCurrentIntegralSpec(AbstractAxesRH, Box):
    """Class for specifying the computation of conduction current via Ampère's circuital law on an axis-aligned loop.

    Example
    -------
    >>> current_spec = AxisAlignedCurrentIntegralSpec(
    ...     center=(0, 0, 0),
    ...     size=(1, 1, 0),
    ...     sign="+",
    ...     snap_contour_to_grid=True
    ... )
    """

    _plane_validator = assert_plane()

    sign: Direction = pd.Field(
        ...,
        title="Direction of Contour Integral",
        description="Positive indicates current flowing in the positive normal axis direction.",
    )

    extrapolate_to_endpoints: bool = pd.Field(
        False,
        title="Extrapolate to Endpoints",
        description="This parameter is passed to :class:`AxisAlignedPathIntegral` objects when computing the contour integral.",
    )

    snap_contour_to_grid: bool = pd.Field(
        False,
        title="Snap Contour to Grid",
        description="This parameter is passed to :class:`AxisAlignedPathIntegral` objects when computing the contour integral.",
    )

    @cached_property
    def main_axis(self) -> Axis:
        """Axis normal to loop"""
        for index, value in enumerate(self.size):
            if value == 0:
                return index

    def _to_path_integral_specs(
        self, h_horizontal=None, h_vertical=None
    ) -> tuple[AxisAlignedPathIntegralSpec, ...]:
        """Returns four ``AxisAlignedPathIntegralSpec`` instances, which represent a contour
        integral around the surface defined by ``self.size``."""
        ax1 = self.remaining_axes[0]
        ax2 = self.remaining_axes[1]

        horizontal_passed = h_horizontal is not None
        vertical_passed = h_vertical is not None
        if self.snap_contour_to_grid and horizontal_passed and vertical_passed:
            (coord1, coord2) = self.remaining_dims

            # Locations where horizontal paths will be snapped
            v_bounds = [
                self.center[ax2] - self.size[ax2] / 2,
                self.center[ax2] + self.size[ax2] / 2,
            ]
            h_snaps = h_horizontal.sel({coord2: v_bounds}, method="nearest").coords[coord2].values
            # Locations where vertical paths will be snapped
            h_bounds = [
                self.center[ax1] - self.size[ax1] / 2,
                self.center[ax1] + self.size[ax1] / 2,
            ]
            v_snaps = h_vertical.sel({coord1: h_bounds}, method="nearest").coords[coord1].values

            bottom_bound = h_snaps[0]
            top_bound = h_snaps[1]
            left_bound = v_snaps[0]
            right_bound = v_snaps[1]
        else:
            bottom_bound = self.bounds[0][ax2]
            top_bound = self.bounds[1][ax2]
            left_bound = self.bounds[0][ax1]
            right_bound = self.bounds[1][ax1]

        # Horizontal paths
        path_size = list(self.size)
        path_size[ax1] = right_bound - left_bound
        path_size[ax2] = 0
        path_center = list(self.center)
        path_center[ax2] = bottom_bound

        bottom = AxisAlignedPathIntegralSpec(
            center=path_center,
            size=path_size,
            extrapolate_to_endpoints=self.extrapolate_to_endpoints,
            snap_path_to_grid=self.snap_contour_to_grid,
        )
        path_center[ax2] = top_bound
        top = AxisAlignedPathIntegralSpec(
            center=path_center,
            size=path_size,
            extrapolate_to_endpoints=self.extrapolate_to_endpoints,
            snap_path_to_grid=self.snap_contour_to_grid,
        )

        # Vertical paths
        path_size = list(self.size)
        path_size[ax1] = 0
        path_size[ax2] = top_bound - bottom_bound
        path_center = list(self.center)

        path_center[ax1] = left_bound
        left = AxisAlignedPathIntegralSpec(
            center=path_center,
            size=path_size,
            extrapolate_to_endpoints=self.extrapolate_to_endpoints,
            snap_path_to_grid=self.snap_contour_to_grid,
        )
        path_center[ax1] = right_bound
        right = AxisAlignedPathIntegralSpec(
            center=path_center,
            size=path_size,
            extrapolate_to_endpoints=self.extrapolate_to_endpoints,
            snap_path_to_grid=self.snap_contour_to_grid,
        )

        return (bottom, right, top, left)

    @add_ax_if_none
    def plot(
        self,
        x: Optional[float] = None,
        y: Optional[float] = None,
        z: Optional[float] = None,
        ax: Ax = None,
        **path_kwargs: Any,
    ) -> Ax:
        """Plot path integral at single (x,y,z) coordinate.

        Parameters
        ----------
        x : float = None
            Position of plane in x direction, only one of x,y,z can be specified to define plane.
        y : float = None
            Position of plane in y direction, only one of x,y,z can be specified to define plane.
        z : float = None
            Position of plane in z direction, only one of x,y,z can be specified to define plane.
        ax : matplotlib.axes._subplots.Axes = None
            Matplotlib axes to plot on, if not specified, one is created.
        **path_kwargs
            Optional keyword arguments passed to the matplotlib plotting of the line.
            For details on accepted values, refer to
            `Matplotlib's documentation <https://tinyurl.com/36marrat>`_.

        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.
        """
        axis, position = self.parse_xyz_kwargs(x=x, y=y, z=z)
        if axis != self.main_axis or not np.isclose(position, self.center[axis], rtol=fp_eps):
            return ax

        plot_params = plot_params_current_path.include_kwargs(**path_kwargs)
        plot_kwargs = plot_params.to_kwargs()
        path_integrals = self._to_path_integral_specs()
        # Plot the path
        for path in path_integrals:
            (xs, ys) = path._vertices_2D(axis)
            ax.plot(xs, ys, **plot_kwargs)

        (ax1, ax2) = self.remaining_axes

        # Add arrow to bottom path, unless right path is longer
        arrow_path = path_integrals[0]
        if self.size[ax2] > self.size[ax1]:
            arrow_path = path_integrals[1]

        (xs, ys) = arrow_path._vertices_2D(axis)
        X = (xs[0] + xs[1]) / 2
        Y = (ys[0] + ys[1]) / 2
        center = np.array([X, Y])
        dx = xs[1] - xs[0]
        dy = ys[1] - ys[0]
        direction = np.array([dx, dy])
        segment_length = np.linalg.norm(direction)
        unit_dir = direction / segment_length

        # Change direction of arrow depending on sign of current definition
        if self.sign == "-":
            unit_dir *= -1.0
        # Change direction of arrow when the "y" axis is dropped,
        # since the plotted coordinate system will be left-handed (x, z)
        if self.main_axis == 1:
            unit_dir *= -1.0

        start = center - unit_dir * segment_length
        end = center
        ax.annotate(
            "",
            xytext=(start[0], start[1]),
            xy=(end[0], end[1]),
            arrowprops=ARROW_CURRENT,
        )
        return ax


class Custom2DCurrentIntegralSpec(Custom2DPathIntegralSpec):
    """Class for specifying the computation of conduction current via Ampère's circuital law on a custom path.
    To compute the current flowing in the positive ``axis`` direction, the vertices should be
    ordered in a counterclockwise direction.

    Example
    -------
    >>> import numpy as np
    >>> vertices = np.array([[0, 0], [2, 0], [2, 1], [0, 1], [0, 0]])
    >>> current_spec = Custom2DCurrentIntegralSpec(
    ...     axis=2,
    ...     position=0,
    ...     vertices=vertices
    ... )
    """

    @add_ax_if_none
    def plot(
        self,
        x: Optional[float] = None,
        y: Optional[float] = None,
        z: Optional[float] = None,
        ax: Ax = None,
        **path_kwargs: Any,
    ) -> Ax:
        """Plot path integral at single (x,y,z) coordinate.

        Parameters
        ----------
        x : float = None
            Position of plane in x direction, only one of x,y,z can be specified to define plane.
        y : float = None
            Position of plane in y direction, only one of x,y,z can be specified to define plane.
        z : float = None
            Position of plane in z direction, only one of x,y,z can be specified to define plane.
        ax : matplotlib.axes._subplots.Axes = None
            Matplotlib axes to plot on, if not specified, one is created.
        **path_kwargs
            Optional keyword arguments passed to the matplotlib plotting of the line.
            For details on accepted values, refer to
            `Matplotlib's documentation <https://tinyurl.com/36marrat>`_.

        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.
        """
        axis, position = Geometry.parse_xyz_kwargs(x=x, y=y, z=z)
        if axis != self.main_axis or not np.isclose(position, self.position, rtol=fp_eps):
            return ax

        plot_params = plot_params_current_path.include_kwargs(**path_kwargs)
        plot_kwargs = plot_params.to_kwargs()
        xs = self.vertices[:, 0]
        ys = self.vertices[:, 1]
        ax.plot(xs, ys, **plot_kwargs)

        # Add arrow at start of contour
        ax.annotate(
            "",
            xytext=(xs[0], ys[0]),
            xy=(xs[1], ys[1]),
            arrowprops=ARROW_CURRENT,
        )
        return ax


class CompositeCurrentIntegralSpec(MicrowaveBaseModel):
    """Specification for a composite current integral.

    Notes
    -----
        This class is used to set up a ``CompositeCurrentIntegral``, which combines
        multiple current integrals. It does not perform any integration itself.

    Example
    -------
    >>> spec1 = AxisAlignedCurrentIntegralSpec(
    ...     center=(0, 0, 0), size=(1, 1, 0), sign="+"
    ... )
    >>> spec2 = AxisAlignedCurrentIntegralSpec(
    ...     center=(2, 0, 0), size=(1, 1, 0), sign="+"
    ... )
    >>> composite_spec = CompositeCurrentIntegralSpec(
    ...     path_specs=(spec1, spec2),
    ...     sum_spec="sum"
    ... )
    """

    path_specs: tuple[Union[AxisAlignedCurrentIntegralSpec, Custom2DCurrentIntegralSpec], ...] = (
        pd.Field(
            ...,
            title="Path Specifications",
            description="Definition of the disjoint path specifications for each isolated contour integral.",
        )
    )

    sum_spec: Literal["sum", "split"] = pd.Field(
        ...,
        title="Sum Specification",
        description="Determines the method used to combine the currents calculated by the different "
        "current integrals defined by ``path_specs``. ``sum`` simply adds all currents, while ``split`` "
        "keeps contributions with opposite phase separate, which allows for isolating the current "
        "flowing in opposite directions. In ``split`` version, the current returned is the maximum "
        "of the two contributions.",
    )

    @add_ax_if_none
    def plot(
        self,
        x: Optional[float] = None,
        y: Optional[float] = None,
        z: Optional[float] = None,
        ax: Ax = None,
        **path_kwargs: Any,
    ) -> Ax:
        """Plot path integral at single (x,y,z) coordinate.

        Parameters
        ----------
        x : float = None
            Position of plane in x direction, only one of x,y,z can be specified to define plane.
        y : float = None
            Position of plane in y direction, only one of x,y,z can be specified to define plane.
        z : float = None
            Position of plane in z direction, only one of x,y,z can be specified to define plane.
        ax : matplotlib.axes._subplots.Axes = None
            Matplotlib axes to plot on, if not specified, one is created.
        **path_kwargs
            Optional keyword arguments passed to the matplotlib plotting of the line.
            For details on accepted values, refer to
            `Matplotlib's documentation <https://tinyurl.com/36marrat>`_.

        Returns
        -------
        matplotlib.axes._subplots.Axes
            The supplied or created matplotlib axes.
        """
        for path_spec in self.path_specs:
            ax = path_spec.plot(x=x, y=y, z=z, ax=ax, **path_kwargs)
        return ax

    @pd.validator("path_specs", always=True)
    def _path_specs_not_empty(cls, val):
        """Makes sure at least one path spec has been supplied"""
        # overall shape of vertices
        if len(val) < 1:
            raise SetupError(
                "'CompositeCurrentIntegralSpec.path_specs' must be a list of one or more current integrals. "
            )
        return val

    @cached_property
    def bounds(self) -> Bound:
        """Return the overall bounding box of all path specifications.

        Computed by taking the union of bounds from all path specs.

        Returns
        -------
        Bound
            Tuple of (rmin, rmax) where rmin and rmax are tuples of (x, y, z) coordinates
            representing the minimum and maximum corners of the bounding box.
        """
        # Start with bounds of first path spec
        overall_bounds = self.path_specs[0].bounds

        # Union with bounds of remaining path specs
        for path_spec in self.path_specs[1:]:
            overall_bounds = bounds_union(overall_bounds, path_spec.bounds)

        return overall_bounds



"""Frequency utilities class ``FrequencyUtils`` and utility class ``FreqRange`` for frequency and wavelength handling."""

from __future__ import annotations

from typing import Any

import numpy as np
import pydantic as pd
import pydantic.v1 as pydantic
from numpy.typing import NDArray

from tidy3d.components.base import Tidy3dBaseModel
from tidy3d.components.source.time import GaussianPulse
from tidy3d.constants import C_0

O_BAND = (1.260, 1.360)
E_BAND = (1.360, 1.460)
S_BAND = (1.460, 1.530)
C_BAND = (1.530, 1.565)
L_BAND = (1.565, 1.625)
U_BAND = (1.625, 1.675)


class FrequencyUtils(Tidy3dBaseModel):
    """Utilities for classifying frequencies/wavelengths and generating samples for standard optical bands."""

    use_wavelength: bool = pd.Field(
        False,
        title="Use wavelength",
        description="Indicate whether to use wavelengths instead of frequencies for the return "
        "values of functions and parameters.",
    )

    def classification(self, value: float) -> tuple[str]:
        """Band classification for a given frequency/wavelength.

        Frequency values must be given in hertz (Hz). Wavelengths must be
        given in micrometers (μm).

        Parameters
        ----------
        value : float
            Value to classify.

        Returns
        -------
        tuple[str]
            String tuple with classification.
        """
        if self.use_wavelength:
            value = C_0 / value
        if value < 3:
            return ("near static",)
        if value < 300e6:
            if value < 30:
                return ("radio wave", "ELF")
            if value < 300:
                return ("radio wave", "SLF")
            if value < 3e3:
                return ("radio wave", "ULF")
            if value < 30e3:
                return ("radio wave", "VLF")
            if value < 300e3:
                return ("radio wave", "LF")
            if value < 3e6:
                return ("radio wave", "MF")
            if value < 30e6:
                return ("radio wave", "HF")
            return ("radio wave", "VHF")
        if value < 300e9:
            if value < 3e9:
                return ("microwave", "UHF")
            if value < 30e9:
                return ("microwave", "SHF")
            return ("microwave", "EHF")
        if value < 400e12:
            if value < 6e12:
                return ("infrared", "FIR")
            if value < 100e12:
                return ("infrared", "MIR")
            return ("infrared", "NIR")
        if value < 790e12:
            if value < 480e12:
                return ("visible", "red")
            if value < 510e12:
                return ("visible", "orange")
            if value < 530e12:
                return ("visible", "yellow")
            if value < 600e12:
                return ("visible", "green")
            if value < 620e12:
                return ("visible", "cyan")
            if value < 670e12:
                return ("visible", "blue")
            return ("visible", "violet")
        if value < 30e15:
            if value < 1e15:
                return ("ultraviolet", "NUV")
            if value < 1.5e15:
                return ("ultraviolet", "MUV")
            if value < 2.47e15:
                return ("ultraviolet", "FUV")
            return ("ultraviolet", "EUV")
        if value < 30e18:
            if value < 3e18:
                return ("X-ray", "soft X-ray")
            return ("X-ray", "hard X-ray")
        return ("γ-ray",)

    def o_band(self, n: int = 11) -> list[float]:
        """
        Optical O band frequencies/wavelengths sorted by wavelength.

        The returned samples are equally spaced in wavelength.

        Parameters
        ----------
        n : int
            Desired number of samples.

        Returns
        -------
        list[float]
            Samples list.
        """
        values = np.linspace(*O_BAND, n)
        if not self.use_wavelength:
            values = C_0 / values
        return values.tolist()

    def e_band(self, n: int = 11) -> list[float]:
        """
        Optical E band frequencies/wavelengths sorted by wavelength.

        The returned samples are equally spaced in wavelength.

        Parameters
        ----------
        n : int
            Desired number of samples.

        Returns
        -------
        list[float]
            Samples list.
        """
        values = np.linspace(*E_BAND, n)
        if not self.use_wavelength:
            values = C_0 / values
        return values.tolist()

    def s_band(self, n: int = 15) -> list[float]:
        """
        Optical S band frequencies/wavelengths sorted by wavelength.

        The returned samples are equally spaced in wavelength.

        Parameters
        ----------
        n : int
            Desired number of samples.

        Returns
        -------
        list[float]
            Samples list.
        """
        values = np.linspace(*S_BAND, n)
        if not self.use_wavelength:
            values = C_0 / values
        return values.tolist()

    def c_band(self, n: int = 8) -> list[float]:
        """
        Optical C band frequencies/wavelengths sorted by wavelength.

        The returned samples are equally spaced in wavelength.

        Parameters
        ----------
        n : int
            Desired number of samples.

        Returns
        -------
        list[float]
            Samples list.
        """
        values = np.linspace(*C_BAND, n)
        if not self.use_wavelength:
            values = C_0 / values
        return values.tolist()

    def l_band(self, n: int = 13) -> list[float]:
        """
        Optical L band frequencies/wavelengths sorted by wavelength.

        The returned samples are equally spaced in wavelength.

        Parameters
        ----------
        n : int
            Desired number of samples.

        Returns
        -------
        list[float]
            Samples list.
        """
        values = np.linspace(*L_BAND, n)
        if not self.use_wavelength:
            values = C_0 / values
        return values.tolist()

    def u_band(self, n: int = 11) -> list[float]:
        """
        Optical U band frequencies/wavelengths sorted by wavelength.

        The returned samples are equally spaced in wavelength.

        Parameters
        ----------
        n : int
            Desired number of samples.

        Returns
        -------
        list[float]
            Samples list.
        """
        values = np.linspace(*U_BAND, n)
        if not self.use_wavelength:
            values = C_0 / values
        return values.tolist()


frequencies = FrequencyUtils(use_wavelength=False)
wavelengths = FrequencyUtils(use_wavelength=True)

frequencies.__doc__ = (
    "Frequency utilities configured to interpret and return values in hertz (Hz). "
    "Use for RF, microwave, optical, and other band classifications in frequency units."
)

wavelengths.__doc__ = (
    "Frequency utilities configured to interpret and return values in micrometers (μm). "
    "Use for optical and photonic calculations where wavelength units are preferred."
)


class FreqRange(Tidy3dBaseModel):
    """
    Convenience class for handling frequency/wavelength conversion; it simplifies specification
    of frequency ranges and sample points for sources and monitors.

    Notes
    -----
        Depending on the context the user can define desired frequency range by specifying:
        - central frequency ``freq0`` and frequency bandwidth ``fwidth``;
        - frequency interval [``fmin``,``fmax``];
        - central wavelength ``wvl0`` and wavelength range ``wvl_width``;
        - wavelength interval [``wvl_min``, ``wvl_max``].

    Example
    -------
    >>> import tidy3d as td
    >>> freq0  = 1e12
    >>> fwidth = 1e11
    >>> freq_range = td.FreqRange(freq0=freq0, fwidth=fwidth)
    >>> central_freq = freq_range.freqs(num_points=1, spacing="uniform_freq")
    >>> freqs = freq_range.freqs(num_points=11, spacing="uniform_wvl")
    >>> source = freq_range.to_gaussian_pulse()
    """

    freq0: pydantic.PositiveFloat = pydantic.Field(
        ...,
        title="Central frequency",
        description="Real-valued positive central frequency.",
        units="Hz",
    )

    fwidth: pydantic.PositiveFloat = pydantic.Field(
        ...,
        title="Frequency bandwidth",
        description="Real-valued positive width of the frequency range (bandwidth).",
        units="Hz",
    )

    @pydantic.root_validator
    def check_half_fwidth_less_than_freq0(cls, values):
        freq0 = values.get("freq0")
        fwidth = values.get("fwidth")
        if freq0 is not None and fwidth is not None:
            if (fwidth / 2) >= freq0:
                raise ValueError(
                    "Frequency bandwidth `fwidth` must be strictly less than `2 * freq0`."
                )
        return values

    @property
    def fmin(self) -> float:
        """Infer lowest frequency ``fmin`` from central frequency ``freq0`` and bandwidth ``fwidth``."""
        return self.freq0 - self.fwidth / 2.0

    @property
    def fmax(self) -> float:
        """Infer highest frequency ``fmax`` from central frequency ``freq0`` and bandwidth ``fwidth``."""
        return self.freq0 + self.fwidth / 2.0

    @property
    def wvl0(self) -> float:
        """Get central wavelength from central frequency and bandwidth."""
        wvl_min = C_0 / self.fmax
        wvl_max = C_0 / self.fmin
        return 0.5 * (wvl_min + wvl_max)

    @classmethod
    def from_freq_interval(cls, fmin: float, fmax: float) -> FreqRange:
        """
        method ``from_freq_interval()`` creates instance of class ``FreqRange`` from frequency interval
        defined by arguments  ``fmin`` and ``fmax``.

        NB: central frequency never corresponds to central wavelength!
        ``freq0 = (fmin + fmax) / 2`` implies that ``wvl0 != (wvl_min + wvl_max) / 2`` and vise-versa.

        Parameters
        ----------
        fmin : float
            Lower bound of frequency of interest.
        fmax : float
            Upper bound of frequency of interest.

        Returns
        -------
        FreqRange
            An instance of ``FreqRange`` defined by frequency interval [``fmin``, ``fmax``].
        """

        # extract frequency-related info
        freq0 = 0.5 * (fmax + fmin)  # extract central freq
        fwidth = fmax - fmin  # extract bandwidth
        return cls(freq0=freq0, fwidth=fwidth)

    @classmethod
    def from_wvl(cls, wvl0: float, wvl_width: float) -> FreqRange:
        """
        method ``from_wvl()`` updated instance of class ``FreqRange`` by reassigning new
        frequency- and wavelength-related parameters.

        NB: central frequency never corresponds to central wavelength!
        ``wvl0 = (wvl_min + wvl_max) / 2`` implies that ``freq0 != (fmin + fmax) / 2`` and vise versa.

        Parameters
        ----------
        wvl0 : float
            Real-valued central wavelength.
        wvl_width : float
            Real-valued wavelength range.

        Returns
        -------
        FreqRange
            An instance of ``FreqRange`` defined by central wavelength ``wvl0`` and wavelength range ``wvl_width``.
        """

        # calculate lowest and highest frequencies
        fmin = C_0 / (wvl0 + wvl_width / 2.0)
        fmax = C_0 / (wvl0 - wvl_width / 2.0)

        return cls.from_freq_interval(fmin=fmin, fmax=fmax)

    @classmethod
    def from_wvl_interval(cls, wvl_min: float, wvl_max: float) -> FreqRange:
        """
        method ``from_wvl_interval()`` updated instance of class ``FreqRange`` by reassigning new
        frequency- and wavelength-related parameters.

        NB: central frequency never corresponds to central wavelength!
        ``wvl0 = (wvl_min + wvl_max) / 2`` implies that ``freq0 != (fmin + fmax) / 2``.

        Parameters
        ----------
        wvl_min : float
            The lowest wavelength of interest.
        wvl_max : float
            The longest wavelength of interest.

        Returns
        -------
        FreqRange
            An instance of ``FreqRange`` defined by the wavelength interval [``wvl_min``, ``wvl_max``].
        """

        # convert wavelength intervals to frequency interval
        fmax = C_0 / wvl_min
        fmin = C_0 / wvl_max

        return cls.from_freq_interval(fmin=fmin, fmax=fmax)

    def freqs(self, num_points: int, spacing: str = "uniform_freq") -> NDArray[np.float64]:
        """
        method ``freqs()`` returns a numpy array of ``num_point`` frequencies uniformly
        sampled from the specified frequency range;
        if ``num_points == 1`` method returns the central frequency ``freq0``.

        Parameters
        ----------
        num_points : int
            Number of frequency points in a frequency range of interest.
        spacing: str = "uniform_freq"
            Mode of frequency sampling.
            If spacing is set to ``uniform_freq``, frequencies are sampled uniformly over the frequency interval.
            If set to ``uniform_wvl``, frequencies correspond to uniformly sampled wavelengths over the wavelength interval.
            Frequencies are sorted in ascending order.

        Returns
        -------
        np.ndarray
            a numpy array of uniformly distributed frequency samples in a frequency range of interest.
        """
        if spacing == "uniform_wvl":
            return (
                np.array([C_0 / self.wvl0])
                if num_points == 1
                else np.sort(C_0 / self.wvls(num_points=num_points, spacing=spacing))
            )

        elif spacing == "uniform_freq":
            return (
                np.array([self.freq0])
                if num_points == 1
                else np.linspace(self.fmin, self.fmax, num_points)
            )

        else:
            raise ValueError(
                "Invalid `spacing` value in FreqRange.freqs(): expected 'uniform_freq' or 'uniform_wvl'. "
                f"Received: {spacing!r}. Please provide a valid spacing type."
            )

    def sweep_decade(self, num_points_per_decade: int) -> NDArray[np.float64]:
        """
        Generate frequencies with logarithmic spacing across decades.

        Notes
        -----
        This method creates a logarithmically spaced frequency sweep.
        It is analogous to the SPICE AC analysis command for a decade
        sweep:  ``.ac dec <num_points_per_decade> <fmin> <fmax>``

        Parameters
        ----------
        num_points_per_decade : int
            Number of frequency points per decade. Must be strictly positive.

        Returns
        -------
        NDArray[np.float64]
            Array of frequencies with logarithmic spacing across decades.

        Raises
        ------
        ValueError
            If ``num_points_per_decade`` is not positive, or if frequency range is invalid.

        Examples
        --------
        >>> import tidy3d as td
        >>> freq_range = td.FreqRange.from_freq_interval(1e3, 1e6)  # 1 kHz to 1 MHz
        >>> freqs = freq_range.sweep_decade(10)  # 10 points per decade
        """
        # Input validation
        if num_points_per_decade <= 0:
            raise ValueError(
                f"'num_points_per_decade' must be strictly positive, got {num_points_per_decade}."
            )

        # Calculate logarithmic range
        fstart = np.log10(self.fmin)
        fend = np.log10(self.fmax)
        num_decades = fend - fstart

        # Calculate total number of points
        # Add 1 to include the endpoint, ensuring we cover the full range
        num_total_points = int(np.round(num_decades * num_points_per_decade)) + 1

        # Generate logarithmically spaced frequencies
        return np.logspace(fstart, fend, num_total_points)

    def wvls(self, num_points: int, spacing: str = "uniform_wvl") -> NDArray[np.float64]:
        """
        method ``wvls()`` returns a numpy array of ``num_points`` wavelengths uniformly
        sampled from the range of wavelengths;
        if ``num_points == 1`` the method returns central wavelength ``wvl0``.

        Parameters
        ----------
        num_points : int
            Number of wavelength points in a range of wavelengths of interest.
        spacing: str = "uniform_wvl"
            Mode of wavelength sampling.
            If spacing is set to ``uniform_wvl``, wavelengths are sampled uniformly over the wavelength interval.
            If set to ``uniform_freq``, wavelengths correspond to uniformly sampled frequencies over the frequency interval.
            Wavelengths are sorted in ascending order.

        Returns
        -------
        np.ndarray
            a numpy array of uniformly distributed wavelength samples in ascending order.
        """
        if spacing == "uniform_freq":
            if num_points == 1:
                return np.array([C_0 / self.freq0])
            else:
                return np.sort(C_0 / self.freqs(num_points=num_points, spacing=spacing))

        elif spacing == "uniform_wvl":
            if num_points == 1:
                return np.array([self.wvl0])
            else:
                wvl_min = C_0 / self.fmax
                wvl_max = C_0 / self.fmin
                return np.linspace(wvl_min, wvl_max, num_points)

        else:
            raise ValueError(
                "Invalid `spacing` value in FreqRange.wvls(): expected 'uniform_freq' or 'uniform_wvl'. "
                f"Received: {spacing!r}. Please provide a valid spacing type."
            )

    def to_gaussian_pulse(self, **kwargs: Any) -> GaussianPulse:
        """
        to_gaussian_pulse(): Return a ``GaussianPulse`` instance based on this frequency range.

        This method constructs a ``GaussianPulse`` using
        ``GaussianPulse.from_frequency_range(fmin=self.fmin, fmax=self.fmax, **kwargs)``.

        If you prefer to define the pulse using ``freq0`` and ``fwidth`` directly,
        you should manually instantiate the ``GaussianPulse`` instead.

        Parameters
        ----------
        kwargs : dict
            Keyword arguments passed to ``GaussianPulse()``, excluding ``freq0`` & ``fwidth``.

        Returns
        -------
        GaussianPulse
            A ``GaussianPulse`` that maximizes its amplitude in the frequency range [``fmin``, ``fmax``].
        """

        duplicate_keys = {"fmin", "fmax"} & kwargs.keys()
        if duplicate_keys:
            is_plural = len(duplicate_keys) > 1
            keys_str = ", ".join(f"'{key}'" for key in sorted(duplicate_keys, reverse=True))
            raise ValueError(
                f"Keyword argument{'s' if is_plural else ''} {keys_str} "
                f"conflict{'' if is_plural else 's'} with values already set in the 'FreqRange' object. "
                f"Please exclude {'them' if is_plural else 'it'} from the 'to_gaussian_pulse()' call."
            )

        # create an instance of GaussianPulse class with defined frequency params
        return GaussianPulse.from_frequency_range(fmin=self.fmin, fmax=self.fmax, **kwargs)



from __future__ import annotations

import autograd.numpy as np
from numpy.typing import NDArray

from tidy3d.plugins.autograd.constants import BETA_DEFAULT, ETA_DEFAULT


def ramp_projection(array: NDArray, width: float = 0.1, center: float = 0.5) -> NDArray:
    """Apply a piecewise linear ramp projection to an array.

    This function performs a ramp projection on the input array, modifying its values
    based on the specified width and center. Values within the range
    [center - width/2, center + width/2] are linearly transformed, while values
    outside this range are projected to 0 or 1. The input and output is assumed to be
    within the range [0, 1].

    Parameters
    ----------
    array : np.ndarray
        The input array to be projected.
    width : float = 0.1
        The width of the ramp.
    center : float 0.5
        The center of the ramp.

    Returns
    -------
    np.ndarray
        The array after applying the ramp projection.
    """
    ll = array <= (center - width / 2)
    cc = (array > (center - width / 2)) & (array < (center + width / 2))
    rr = array >= (center + width / 2)

    return np.concatenate(
        [
            np.zeros(array[ll].size),
            (array[cc] - (center - width / 2)) / width,
            np.ones(array[rr].size),
        ]
    )


def tanh_projection(
    array: NDArray, beta: float = BETA_DEFAULT, eta: float = ETA_DEFAULT
) -> NDArray:
    """Apply a tanh-based soft-thresholding projection to an array.

    This function performs a tanh projection on the input array, which is a common
    soft-thresholding scheme used in topology optimization. The projection modifies
    the values of the array based on the specified `beta` and `eta` parameters.

    Parameters
    ----------
    array : np.ndarray
        The input array to be projected.
    beta : float = BETA_DEFAULT
        The steepness of the projection. Higher values result in a sharper transition.
    eta : float = ETA_DEFAULT
        The midpoint of the projection.

    Returns
    -------
    np.ndarray
        The array after applying the tanh projection.
    """
    if beta == 0:
        return array
    if beta == np.inf:
        return np.where(array > eta, 1.0, 0.0)
    num = np.tanh(beta * eta) + np.tanh(beta * (array - eta))
    denom = np.tanh(beta * eta) + np.tanh(beta * (1 - eta))
    return num / denom


def smoothed_projection(
    array: NDArray,
    beta: float = BETA_DEFAULT,
    eta: float = ETA_DEFAULT,
    scaling_factor=1.0,
) -> NDArray:
    """
    Apply a subpixel-smoothed projection method.
    The subpixel-smoothed projection method is discussed in [1]_ as follows:

    This projection method eliminates discontinuities by applying first-order
    smoothing at material boundaries through analytical fill factors. Unlike
    traditional quadrature approaches, it works with maximum projection strength
    (:math:`\\beta = \\infty`) and derives closed-form expressions for interfacial regions.

    Prerequisites: input fields must be pre-filtered for continuity (for example
    using a conic filter).

    The algorithm detects whether boundaries intersect grid cells. When interfaces
    are absent, standard projection is applied. For cells containing boundaries,
    analytical fill ratios are computed to maintain gradient continuity as interfaces
    move through cells and traverse pixel centers. This enables arbitrarily large
    :math:`\\beta` values while preserving differentiability throughout the transition
    process.

    .. warning::
        This function assumes that the device is placed on a uniform grid. When using
        ```GridSpec.auto``` in the simulation, make sure to place a ``MeshOverrideStructure`` at
        the position of the optimized geometry.

    Parameters
    ----------
    array : np.ndarray
        The input array to be projected.
    beta : float = BETA_DEFAULT
        The steepness of the projection. Higher values result in a sharper transition.
    eta : float = ETA_DEFAULT
        The midpoint of the projection.
    scaling_factor: float = 1.0
        Optional scaling factor to adjust dx and dy to different resolutions.

    Example
    -------
        >>> import autograd.numpy as np
        >>> from tidy3d.plugins.autograd.invdes.filters import ConicFilter
        >>> arr = np.random.uniform(size=(50, 50))
        >>> filter = ConicFilter(kernel_size=5)
        >>> arr_filtered = filter(arr)
        >>> eta = 0.5  # center of projection
        >>> smoothed = smoothed_projection(arr_filtered, beta=np.inf, eta=eta)

    .. [1] A. M. Hammond, A. Oskooi, I. M. Hammond, M. Chen, S. E. Ralph, and
       S. G. Johnson, "Unifying and accelerating level-set and density-based topology
       optimization by subpixel-smoothed projection," arXiv:2503.20189v3 [physics.optics]
       (2025).
    """
    # sanity checks
    if array.ndim != 2:
        raise ValueError(f"Smoothed projection expects a 2d-array, but got shape {array.shape=}")

    # smoothing kernel is circle (or ellipse for non-uniform grid)
    # we choose smoothing kernel with unit area, which is r~=0.56, a bit larger than (arbitrary) default r=0.55 in paper
    dx = dy = scaling_factor
    smooth_radius = np.sqrt(1 / np.pi) * scaling_factor

    original_projected = tanh_projection(array, beta=beta, eta=eta)

    # finite-difference spatial gradients
    rho_filtered_grad = np.gradient(array)
    rho_filtered_grad_helper = (rho_filtered_grad[0] / dx) ** 2 + (rho_filtered_grad[1] / dy) ** 2

    nonzero_norm = np.abs(rho_filtered_grad_helper) > 1e-10

    filtered_grad_norm = np.sqrt(np.where(nonzero_norm, rho_filtered_grad_helper, 1))
    filtered_grad_norm_eff = np.where(nonzero_norm, filtered_grad_norm, 1)

    # distance of pixel center to nearest interface
    distance = (eta - array) / filtered_grad_norm_eff

    needs_smoothing = nonzero_norm & (np.abs(distance) < smooth_radius)

    # double where trick
    d_rel = distance / smooth_radius
    polynom = np.where(
        needs_smoothing, 0.5 - 15 / 16 * d_rel + 5 / 8 * d_rel**3 - 3 / 16 * d_rel**5, 1.0
    )
    # F(-d)
    polynom_neg = np.where(
        needs_smoothing, 0.5 + 15 / 16 * d_rel - 5 / 8 * d_rel**3 + 3 / 16 * d_rel**5, 1.0
    )

    # two projections, one for lower and one for upper bound
    rho_filtered_minus = array - smooth_radius * filtered_grad_norm_eff * polynom
    rho_filtered_plus = array + smooth_radius * filtered_grad_norm_eff * polynom_neg
    rho_minus_eff_projected = tanh_projection(rho_filtered_minus, beta=beta, eta=eta)
    rho_plus_eff_projected = tanh_projection(rho_filtered_plus, beta=beta, eta=eta)

    # Smoothing is only applied to projections
    projected_smoothed = (1 - polynom) * rho_minus_eff_projected + polynom * rho_plus_eff_projected
    smoothed = np.where(
        needs_smoothing,
        projected_smoothed,
        original_projected,
    )
    return smoothed



from __future__ import annotations

from collections.abc import Iterable, Sequence
from functools import cache

import autograd.numpy as anp
import numpy as np
import scipy.ndimage
from autograd.extend import defjvp, defvjp, primitive


def _normalize_sequence(value: float | Sequence[float], ndim: int) -> tuple[float, ...]:
    """Convert a scalar or sequence into a tuple of length ``ndim``."""
    if isinstance(value, Iterable) and not np.isscalar(value):
        value_tuple = tuple(value)
        if len(value_tuple) != ndim:
            raise ValueError(f"Sequence length {len(value_tuple)} does not match ndim {ndim}.")
        return tuple(float(v) for v in value_tuple)
    return (float(value),) * ndim


def _normalize_modes(mode: str | Sequence[str], ndim: int) -> tuple[str, ...]:
    """Normalize a padding mode argument into a tuple of strings with length ``ndim``."""
    if isinstance(mode, str):
        return (mode,) * ndim
    mode_tuple = tuple(mode)
    if len(mode_tuple) != ndim:
        raise ValueError(f"Mode sequence length {len(mode_tuple)} does not match ndim {ndim}.")
    return mode_tuple


@cache
def _gaussian_weight_matrix(
    length: int,
    sigma: float,
    mode: str,
    truncate: float,
    order: int,
    cval: float,
) -> np.ndarray:
    """Return the 1-D Gaussian filter matrix used along a single axis."""
    if sigma <= 0.0:
        return np.eye(length)
    eye = np.eye(length, dtype=float)
    weights = scipy.ndimage.gaussian_filter1d(
        eye,
        sigma=sigma,
        axis=0,
        order=order,
        mode=mode,
        cval=cval,
        truncate=truncate,
    )
    return weights


@primitive
def gaussian_filter(
    array,
    sigma,
    *,
    order=0,
    mode="reflect",
    cval=0.0,
    truncate=4.0,
    **kwargs,
):
    return scipy.ndimage.gaussian_filter(
        array,
        sigma,
        order=order,
        mode=mode,
        cval=cval,
        truncate=truncate,
        **kwargs,
    )


def _gaussian_filter_vjp(
    ans, array, sigma, *, order=0, mode="reflect", cval=0.0, truncate=4.0, **kwargs
):
    ndim = array.ndim
    sigma_seq = _normalize_sequence(sigma, ndim)
    order_seq = _normalize_sequence(order, ndim)
    truncate_seq = _normalize_sequence(truncate, ndim)
    cval_seq = _normalize_sequence(cval, ndim)
    mode_seq = _normalize_modes(mode, ndim)

    if any(int(o) != 0 for o in order_seq):
        raise NotImplementedError("gaussian_filter VJP currently supports only order=0.")
    if kwargs:
        raise NotImplementedError(
            f"gaussian_filter VJP does not support additional keyword arguments: {tuple(kwargs)}"
        )

    def vjp(g):
        grad = np.asarray(g)
        for axis in reversed(range(ndim)):
            sigma_axis = float(sigma_seq[axis])
            if sigma_axis <= 0.0:
                continue
            mode_axis = mode_seq[axis]
            truncate_axis = float(truncate_seq[axis])
            order_axis = int(order_seq[axis])
            cval_axis = float(cval_seq[axis])
            length = grad.shape[axis]
            weights = _gaussian_weight_matrix(
                length, sigma_axis, mode_axis, truncate_axis, order_axis, cval_axis
            )
            grad = np.tensordot(weights.T, grad, axes=([1], [axis]))
            grad = np.moveaxis(grad, 0, axis)
        return grad

    return vjp


defvjp(gaussian_filter, _gaussian_filter_vjp, argnums=[0])


anp.unwrap = primitive(anp.unwrap)
defjvp(anp.unwrap, lambda g, ans, x, *args, **kwargs: g)
defvjp(anp.unwrap, lambda ans, x, *args, **kwargs: lambda g: g)
